// indexed binary heap priority queue
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "BHPQ.h"

#include <stdio.h>
#include <stdlib.h>

/* move entry up queue O(log n) */
void swap_up(const int index, BinaryHeapPriorityQueue *bhpq);

/* move entry down queue O(log n) */
void swap_down(const int index, BinaryHeapPriorityQueue *bhpq);

/* initialise binary heap priority queue */
BinaryHeapPriorityQueue *bhpq_init(const unsigned int size)
{
	unsigned int i;
	BinaryHeapPriorityQueue *bhpq;

	if(!(bhpq = (BinaryHeapPriorityQueue *) \
			malloc(sizeof(BinaryHeapPriorityQueue)))) {
		fprintf(stderr, \
				"error: couldn't initialise binary heap priority queue\n");
		exit(EXIT_FAILURE);
	}
	bhpq->size = 0;
	bhpq->max_size = size;
	if(!(bhpq->handle = (int *) calloc(bhpq->max_size, sizeof(int)))) {
		fprintf(stderr, \
				"error: couldn't initialise binary heap priority queue\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < bhpq->max_size; i++)
		bhpq->handle[i] = -1;
	bhpq->max_size++;
	if(!(bhpq->data = (QueueEntry **) calloc(bhpq->max_size, \
			sizeof(QueueEntry *)))) {
		fprintf(stderr, \
				"error: couldn't initialise binary heap priority queue\n");
		exit(EXIT_FAILURE);
	}
	for(i = 0; i < bhpq->max_size; i++)
		if(!(bhpq->data[i] = (QueueEntry *) malloc(sizeof(QueueEntry)))) {
			fprintf(stderr, \
					"error: couldn't initialise binary heap priority queue\n");
			exit(EXIT_FAILURE);
		}

	return bhpq;
}

/* free binary heap priority queue */
void bhpq_free(BinaryHeapPriorityQueue *bhpq)
{
	unsigned int i;

	for(i = 0; i < bhpq->max_size; i++)
		free(bhpq->data[i]);
	free(bhpq->data);
	free(bhpq->handle);
	free(bhpq);
}

/* get total size in bytes of binary heap priority queue */
unsigned int bhpq_storage_size(BinaryHeapPriorityQueue *bhpq)
{
	return (sizeof(BinaryHeapPriorityQueue) + bhpq->max_size * sizeof(int) + \
			bhpq->max_size * sizeof(QueueEntry *) + bhpq->max_size * \
			sizeof(QueueEntry));
}

/* insert entry into binary heap priority queue O(log n) */
void bhpq_insert(QueueEntry *entry, BinaryHeapPriorityQueue *bhpq)
{
	unsigned int index;
	int *handle;
	QueueEntry **data;

	data = bhpq->data;
	handle = bhpq->handle;

	/* check if already entry in queue */
	if(handle[entry->key2] >= 0) {
		/* already entry in queue, so reprioritise current entry */
		index = handle[entry->key2];
		if(entry->key1 > data[index]->key1) {
			data[index]->key1 = entry->key1;
			swap_down(index, bhpq);
		} else {
			data[index]->key1 = entry->key1;
			swap_up(index, bhpq);
		}
	} else {
		/* insert new entry */
		index = ++bhpq->size;
		data[index]->key1 = entry->key1;
		data[index]->key2 = entry->key2;
		swap_up(index, bhpq);
	}
}

/* remove entry from binary heap priority queue O(log n) */
int bhpq_remove_head(QueueEntry *entry, BinaryHeapPriorityQueue *bhpq)
{
	unsigned int size;
	int *handle;
	QueueEntry **data;

	if(bhpq->size == 0)
		return 0;

	data = bhpq->data;
	handle = bhpq->handle;
	size = bhpq->size;

	entry->key1 = data[1]->key1;
	entry->key2 = data[1]->key2;
	handle[entry->key2] = -1;

	if(!--bhpq->size)
		return 1;

	data[1]->key1 = data[size]->key1;
	data[1]->key2 = data[size]->key2;

	swap_down(1, bhpq);

	return 1;
}

/* delete entry from binary heap priority queue O(log n) */
void bhpq_delete_head(BinaryHeapPriorityQueue *bhpq)
{
	unsigned int size;
	int *handle;
	QueueEntry **data;

	if(bhpq->size == 0)
		return;

	data = bhpq->data;
	handle = bhpq->handle;
	size = bhpq->size;

	handle[data[1]->key2] = -1;

	if(!--bhpq->size)
		return;

	data[1]->key1 = data[size]->key1;
	data[1]->key2 = data[size]->key2;

	swap_down(1, bhpq);

	return;
}

/* peek at entry at head of queue O(1) */
int bhpq_peek_head(QueueEntry *entry, BinaryHeapPriorityQueue *bhpq)
{
	QueueEntry **data;

	if(bhpq->size == 0)
		return 0;

	data = bhpq->data;

	entry->key1 = data[1]->key1;
	entry->key2 = data[1]->key2;

	return 1;
}

/* replace entry at head of queue O(log n) */
void bhpq_replace_head(QueueEntry *entry, BinaryHeapPriorityQueue *bhpq)
{
	QueueEntry **data;

	data = bhpq->data;

	data[1]->key1 = entry->key1;
	swap_down(1, bhpq);
}

/* move entry up queue */
void swap_up(const int index, BinaryHeapPriorityQueue *bhpq)
{
	unsigned int child, parent, size;
	int *handle;
	QueueEntry **data, entry;
	double key1;

	data = bhpq->data;
	handle = bhpq->handle;
	size = bhpq->size;

	key1 = entry.key1 = data[index]->key1;
	entry.key2 = data[index]->key2;

	child = index;
	parent = child / 2;
	while(parent >= 1 && data[parent]->key1 > key1) {
		data[child]->key1 = data[parent]->key1;
		data[child]->key2 = data[parent]->key2;
		handle[data[child]->key2] = child;
		child = parent;
		parent /= 2;
	}

	data[child]->key1 = key1;
	data[child]->key2 = entry.key2;
	handle[entry.key2] = child;
}

/* move entry down queue */
void swap_down(const int index, BinaryHeapPriorityQueue *bhpq)
{
	unsigned int child, child2, parent, size;
	int *handle;
	QueueEntry **data, entry;
	double key1;

	data = bhpq->data;
	handle = bhpq->handle;
	size = bhpq->size;

	key1 = entry.key1 = data[index]->key1;
	entry.key2 = data[index]->key2;

	parent = index;
	child = 2 * parent;
	child2 = child + 1;
	while(child <= size) {
		if((child2 <= size) && \
				(data[child]->key1 > data[child2]->key1))
			child++;
		if(key1 <= data[child]->key1)
			break;
		data[parent]->key1 = data[child]->key1;
		data[parent]->key2 = data[child]->key2;
		handle[data[parent]->key2] = parent;
		parent = child;
		child = 2 * parent;
		child2 = child + 1;
	}

	data[parent]->key1 = key1;
	data[parent]->key2 = entry.key2;
	handle[data[parent]->key2] = parent;
}
