// hdf5 dataset functions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "Dataset.h"

#include <iostream>
#include <stdlib.h>
#include <string.h>

using namespace ECSB;

// get size of hdf5 datatype
hssize_t ECSB::getDatasetSize(hid_t dset)
{
	hid_t dspace = H5Dget_space(dset);
	hssize_t size = H5Sget_simple_extent_npoints(dspace);
	H5Sclose(dspace);
	return size;
}

// create an hdf5 chunked property list
hid_t ECSB::createChunkedPropertyList(const int rank, hsize_t *dims, \
		const hid_t dtype, unsigned long fill_value)
{
	int i;
	hsize_t total, maxsize = 4294967296LLU;
	bool finished = false;

	// make sure chunk size < maximum allowed chunk size (4GB)
	while(!finished) {
		total = 1;
		for(i = 0; i < rank; i++)
			total *= dims[i];
		total *= H5Tget_size(dtype);
		if(total < maxsize)
			finished = true;
		else {
			dims[rank - 1] /= 2;
			if(dims[rank - 1] == 0)
				break;
		}
	}
	if(!finished) {
		std::cerr << "error: couldn't create chunked property list\n";
		exit(EXIT_FAILURE);
	}

	// create property list
	hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
	if(H5Pset_chunk(cparams, rank, dims) < 0) {
		std::cerr << "error: couldn't create chunked property list\n";
		exit(EXIT_FAILURE);
	}
	if(H5Pset_fill_value(cparams, dtype, &fill_value) < 0) {
		std::cerr << "error: couldn't create chunked property list\n";
		exit(EXIT_FAILURE);
	}

	return cparams;
}

// create an hdf5 file
hid_t ECSB::createDatafile(const char *filename)
{
	hid_t dfile;

	if((dfile = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT)) \
			< 0) {
		std::cerr << "error: couldn't create file " << filename << "\n";
		exit(EXIT_FAILURE);
	}
	
	return dfile;
}

// open an hdf5 file
hid_t ECSB::openDatafile(const char *filename)
{
	hid_t dfile;

	if((dfile = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open file " << filename << "\n";
		exit(EXIT_FAILURE);
	}
	
	return dfile;
}

// close an hdf5 datafile
void ECSB::closeDatafile(hid_t dfile)
{
	if(H5Fclose(dfile) < 0) {
		std::cerr << "error: couldn't close datafile\n";
		exit(EXIT_FAILURE);
	}
}

// create an hdf5 dataspace
hid_t ECSB::createDataspace(const unsigned long size)
{
	hid_t dspace;

	hsize_t dims[1] = { (hsize_t) size };
	hsize_t maxdims[1] = { H5S_UNLIMITED };

	if((dspace = H5Screate_simple(1, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create dataspace\n";
		exit(EXIT_FAILURE);
	}

	return dspace;
}

// create an hdf5 dataspace with chunk size that fits into memory
hid_t ECSB::createDataspaceScaled(unsigned long *size)
{
	hid_t dspace;
	bool finished = false;
	hsize_t dims[1];
	hsize_t maxdims[1] = { H5S_UNLIMITED };
	H5E_auto2_t old_func;
	void *old_client_data;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	while(!finished) {
		dims[0] = (hsize_t) *size;
		if((dspace = H5Screate_simple(1, dims, maxdims)) < 0) {
			*size /= 2;
			if(*size == 0)
				break;
		} else
			finished = true;
	}
	if(!finished) {
		std::cerr << "error: couldn't create dataspace\n";
		exit(EXIT_FAILURE);
	}

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);

	return dspace;
}

// create an hdf5 dataspace
hid_t ECSB::createDataspace(int rank, const hsize_t *dims, \
		const hsize_t *maxdims)
{
	hid_t dspace;

	if((dspace = H5Screate_simple(rank, dims, maxdims)) < 0) {
		std::cerr << "error: couldn't create dataspace\n";
		exit(EXIT_FAILURE);
	}

	return dspace;
}

// create an hdf5 dataspace with chunk size that fits into memory 
hid_t ECSB::createDataspaceScaled(int rank, hsize_t *dims, hsize_t *maxdims)
{
	hid_t dspace;
	bool finished = false;
	H5E_auto2_t old_func;
	void *old_client_data;

	// turn off printing of hdf5 errors
	H5Eget_auto2(H5E_DEFAULT, &old_func, &old_client_data);
	H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

	while(!finished) {
		if((dspace = H5Screate_simple(rank, dims, maxdims)) < 0) {
			dims[rank - 1] /= 2;
			if(dims[rank - 1] == 0)
				break;
		} else
			finished = true;
	}
	if(!finished) {
		std::cerr << "error: couldn't create dataspace\n";
		exit(EXIT_FAILURE);
	}

	// turn back on printing of hdf5 errors
	H5Eset_auto2(H5E_DEFAULT, old_func, old_client_data);

	return dspace;
}

// close an hdf5 dataspace
void ECSB::closeDataspace(hid_t dspace)
{
	if(H5Sclose(dspace) < 0) {
		std::cerr << "error: couldn't close dataspace\n";
		exit(EXIT_FAILURE);
	}
}

// create an hdf5 dataset
hid_t ECSB::createDataset(const char *name, hid_t dfile, hid_t dtype, \
		hid_t dspace)
{
	hid_t dset;

	hsize_t dims[1] = { (hsize_t) H5Sget_simple_extent_npoints(dspace) };
	if(!dims[0])  // handle zero size datasets
		dims[0] = 1;
	hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
	if(H5Pset_chunk(cparams, 1, dims) < 0) {
		std::cerr << "error: couldn't create chunked dataset \"" << name \
				<< "\"\n";
		exit(EXIT_FAILURE);
	}
	if((dset = H5Dcreate2(dfile, name, dtype, dspace, H5P_DEFAULT, cparams, \
			H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't create chunked dataset \"" << name \
				<< "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Pclose(cparams);
	
	return dset;
}

// open an hdf5 dataset
hid_t ECSB::openDataset(const char *name, hid_t dfile)
{
	hid_t dset;

	if((dset = H5Dopen2(dfile, name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}

	return dset;
}

// write an hdf5 dataset
void ECSB::writeDataset(void *entry, void *buffer, hid_t dset, \
		hid_t dtype, hid_t dspace, unsigned long &buffer_index, \
		const unsigned long chunk_size)
{
	hsize_t size = (hsize_t) H5Tget_size(dtype);
	register DS_BYTE *row = (DS_BYTE *) buffer + buffer_index * size;
	memcpy(row, entry, size);
	if(++buffer_index == chunk_size) {
		// write data in buffer to disk
		size = H5Dget_storage_size(dset) / size;
		hsize_t offset[1] = { size };
		hsize_t dims1[1] = { size + chunk_size };
		hsize_t dims2[1] = { chunk_size };
		if(H5Dset_extent(dset, dims1) < 0) {
			char name[512];
			H5Iget_name(dset, name, 512);
			std::cerr << "error: couldn't write dataset \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}
		hid_t fspace = H5Dget_space(dset);
		if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, \
				dims2, NULL) < 0) {
			char name[512];
			H5Iget_name(dset, name, 512);
			std::cerr << "error: couldn't write dataset \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}
		if(H5Dwrite(dset, dtype, dspace, fspace, H5P_DEFAULT, buffer) < 0) {
			char name[512];
			H5Iget_name(dset, name, 512);
			std::cerr << "error: couldn't write dataset \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}
		buffer_index = 0;
		H5Sclose(fspace);
	}
}

// close an hdf5 dataset after reading
void ECSB::closeDataset(hid_t dset)
{
	if(H5Dclose(dset) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't close dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
}

// close an hdf5 dataset after writing
void ECSB::closeDataset(hid_t dfile, hid_t dset, hid_t dtype, \
		void *buffer, unsigned long &buffer_index, \
		const unsigned long chunk_size)
{
	hid_t dspace;

	// create dataspace for remaining entries
	hsize_t size = (hsize_t) H5Dget_storage_size(dset) / H5Tget_size(dtype);
	hsize_t offset[1] = { size };
	hsize_t dims1[1] = { size + buffer_index };
	hsize_t dims2[1] = { buffer_index };
	hsize_t max_dims[1] = { H5S_UNLIMITED };
	if((dspace = H5Screate_simple(1, dims2, max_dims)) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't close dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	if(dims1[0] > 0 && dims1[0] < chunk_size) {
		// resize dataset as smaller than chunk size
		char name[512];
		H5Iget_name(dset, name, 512);
		H5Dclose(dset);
		H5Ldelete(dfile, name, H5P_DEFAULT);
		hid_t cparams = H5Pcreate(H5P_DATASET_CREATE);
		if(H5Pset_chunk(cparams, 1, dims1) < 0) {
			std::cerr << "error: couldn't create dataset \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}
		if((dset = H5Dcreate2(dfile, name, dtype, dspace, H5P_DEFAULT, cparams, \
				H5P_DEFAULT)) < 0) {
			std::cerr << "error: couldn't create dataset \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}
		H5Pclose(cparams);
	} else
		// otherwise extend the dataset
		if(H5Dset_extent(dset, dims1) < 0) {
			char name[512];
			H5Iget_name(dset, name, 512);
			std::cerr << "error: couldn't write dataset \"" << name << "\"\n";
			exit(EXIT_FAILURE);
		}

	// write the remaining entries to file
	hid_t fspace = H5Dget_space(dset);
	if(H5Sselect_hyperslab(fspace, H5S_SELECT_SET, offset, NULL, dims2, \
			NULL) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't write dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	if(H5Dwrite(dset, dtype, dspace, fspace, H5P_DEFAULT, buffer) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't write dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}

	// free resources
	buffer_index = 0;
	H5Sclose(fspace);
	H5Sclose(dspace);
	H5Dclose(dset);
}

// read all entries in an hdf5 dataset in one go
void ECSB::readDataset(void *buffer, hid_t dset)
{
	hid_t dtype = H5Dget_type(dset);
	if(H5Dread(dset, dtype, H5S_ALL, H5S_ALL, H5P_DEFAULT, buffer) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't read dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Tclose(dtype);
}

// read an hdf5 dataset in chunks
unsigned long ECSB::readDataset(void *buffer, hid_t dset, hid_t dtype, \
		unsigned long &chunk_index, const unsigned long chunk_size)
{
	hid_t mspace;

	// create memory space for chunk
	hid_t dspace = H5Dget_space(dset);
	hssize_t size = H5Sget_simple_extent_npoints(dspace);
	hsize_t offset[1] = { chunk_index * chunk_size };
	if(size <= offset[0])
		return 0;  // all chunks read
	hsize_t dims[1] = { size - offset[0] };
	if(dims[0] > chunk_size)
		dims[0] = chunk_size;
	if((mspace = H5Screate_simple(1, dims, NULL)) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't read dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	// select chunk in dataspace
	if(H5Sselect_hyperslab(dspace, H5S_SELECT_SET, offset, NULL, dims, \
			NULL) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't read dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	// read chunk from dataset
	if(H5Dread(dset, dtype, mspace, dspace, H5P_DEFAULT, buffer) < 0) {
		char name[512];
		H5Iget_name(dset, name, 512);
		std::cerr << "error: couldn't read dataset \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	chunk_index++;

	// free resources
	H5Sclose(dspace);
	H5Sclose(mspace);
	
	return dims[0];
}

// create an hdf5 group
hid_t ECSB::createGroup(const char *name, hid_t dfile)
{
	hid_t group;

	if((group = H5Gcreate2(dfile, name, H5P_DEFAULT, H5P_DEFAULT, \
			H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't create group \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}

	return group;
}

// open an hdf5 group
hid_t ECSB::openGroup(const char *name, hid_t dfile)
{
	hid_t group;

	if((group = H5Gopen2(dfile, name, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't open group \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}

	return group;
}

// close an hdf5 group
void ECSB::closeGroup(hid_t group)
{
	if(H5Gclose(group) < 0) {
		std::cerr << "error: couldn't close group\n";
		exit(EXIT_FAILURE);
	}
}

// write an attribute to an hdf5 datafile
void ECSB::writeAttribute(const char *name, const char *value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	hid_t dspace = H5Screate(H5S_SCALAR);
	hid_t dtype = H5Tcopy(H5T_C_S1);
	H5Tset_size(dtype, strlen(value));
	H5Tset_strpad(dtype, H5T_STR_NULLTERM);
	hid_t dattr = H5Acreate2(dgroup, name, dtype, dspace, H5P_DEFAULT, \
			H5P_DEFAULT);
	if(H5Awrite(dattr, dtype, value) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Tclose(dtype);
	H5Sclose(dspace);
	H5Gclose(dgroup);
}

// write an attribute to an hdf5 datafile
void ECSB::writeAttribute(const char *name, const int value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	hid_t dspace = H5Screate(H5S_SCALAR);
	hid_t dattr = H5Acreate2(dgroup, name, H5T_NATIVE_INT, dspace, H5P_DEFAULT, \
			H5P_DEFAULT);
	if(H5Awrite(dattr, H5T_NATIVE_INT, &value) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Sclose(dspace);
	H5Gclose(dgroup);
}

// write an attribute to an hdf5 datafile
void ECSB::writeAttribute(const char *name, const unsigned long value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	hid_t dspace = H5Screate(H5S_SCALAR);
	hid_t dattr = H5Acreate2(dgroup, name, H5T_NATIVE_ULONG, dspace, \
			H5P_DEFAULT, H5P_DEFAULT);
	if(H5Awrite(dattr, H5T_NATIVE_ULONG, &value) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Sclose(dspace);
	H5Gclose(dgroup);
}

// write an attribute to an hdf5 datafile
void ECSB::writeAttribute(const char *name, const unsigned long long value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	hid_t dspace = H5Screate(H5S_SCALAR);
	hid_t dattr = H5Acreate2(dgroup, name, H5T_NATIVE_ULLONG, dspace, \
			H5P_DEFAULT, H5P_DEFAULT);
	if(H5Awrite(dattr, H5T_NATIVE_ULLONG, &value) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Sclose(dspace);
	H5Gclose(dgroup);
}

// write an attribute to an hdf5 datafile
void ECSB::writeAttribute(const char *name, const double value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	hid_t dspace = H5Screate(H5S_SCALAR);
	hid_t dattr = H5Acreate2(dgroup, name, H5T_NATIVE_DOUBLE, dspace, \
			H5P_DEFAULT, H5P_DEFAULT);
	if(H5Awrite(dattr, H5T_NATIVE_DOUBLE, &value) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Sclose(dspace);
	H5Gclose(dgroup);
}

// write an attribute to an hdf5 datafile
void ECSB::writeAttribute(const char *name, const bool value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	hid_t dspace = H5Screate(H5S_SCALAR);
	hid_t dattr = H5Acreate2(dgroup, name, H5T_NATIVE_HBOOL, dspace, \
			H5P_DEFAULT, H5P_DEFAULT);
	if(H5Awrite(dattr, H5T_NATIVE_HBOOL, &value) < 0) {
		std::cerr << "error: couldn't write attribute \"" << name << "\"\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Sclose(dspace);
	H5Gclose(dgroup);
}

// read an attribute from an hdf5 datafile
bool ECSB::readAttribute(const char *name, char *&value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	hid_t dtype = H5Aget_type(dattr);
	if(!(value = (char *) calloc(H5Tget_size(dtype) + 1, sizeof(char)))) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	if(H5Aread(dattr, dtype, value) < 0) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	H5Tclose(dtype);
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// read an attribute from an hdf5 datafile
bool ECSB::readAttribute(const char *name, int &value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	if(H5Aread(dattr, H5T_NATIVE_INT, &value) < 0) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// read an attribute from an hdf5 datafile
bool ECSB::readAttribute(const char *name, unsigned long &value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	if(H5Aread(dattr, H5T_NATIVE_ULONG, &value) < 0) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// read an attribute from an hdf5 datafile
bool ECSB::readAttribute(const char *name, unsigned long long &value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	if(H5Aread(dattr, H5T_NATIVE_ULLONG, &value) < 0) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// read an attribute from an hdf5 datafile
bool ECSB::readAttribute(const char *name, double &value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	if(H5Aread(dattr, H5T_NATIVE_DOUBLE, &value) < 0) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// read an attribute from an hdf5 datafile
bool ECSB::readAttribute(const char *name, bool &value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	if(H5Aread(dattr, H5T_NATIVE_HBOOL, &value) < 0) {
		std::cerr << "error: couldn't read attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// convert an attribute from an hdf5 datafile to a string
bool ECSB::attributeAsString(const char *name, char *&value, \
		hid_t dfile, const char *group)
{
	hid_t dgroup, dattr, dtype, ndtype;
	bool ret;

	if((dgroup = H5Gopen2(dfile, group, H5P_DEFAULT)) < 0)
		return false;
	if((dattr = H5Aopen(dgroup, name, H5P_DEFAULT)) < 0) {
		H5Gclose(dgroup);
		return false;
	}
	if(!(value = (char *) calloc(1024, sizeof(char)))) {
		std::cerr << "error: couldn't convert attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
	dtype = H5Aget_type(dattr);
	ndtype = H5Tget_class(dtype);
	ret = false;
	if(ndtype == H5T_INTEGER) {
		long int aval = 0;
		if(H5Aread(dattr, dtype, &aval) < 0) {
			std::cerr << "error: couldn't convert attribute " << name << "\n";
			exit(EXIT_FAILURE);
		}
		snprintf(value, 1024, "%ld", (long int) aval);
		ret = true;
	}
	if(ndtype == H5T_FLOAT) {
		double aval;
		if(H5Aread(dattr, dtype, &aval) < 0) {
			std::cerr << "error: couldn't convert attribute " << name << "\n";
			exit(EXIT_FAILURE);
		}
		snprintf(value, 1024, "%f", aval);
		ret = true;
	}
	if(ndtype == H5T_STRING) {
		if(H5Aread(dattr, dtype, value) < 0) {
			std::cerr << "error: couldn't convert attribute " << name << "\n";
			exit(EXIT_FAILURE);
		}
		ret = true;
	}
	if(!ret) {
		std::cerr << "error: couldn't convert attribute " << name << "\n";
		exit(EXIT_FAILURE);
	}
/*	if(ndtype == H5T_NATIVE_ULONG) {
		unsigned long aval;
		if(H5Aread(dattr, dtype, &aval) < 0) {
			std::cerr << "error: couldn't convert attribute " << name << "\n";
			exit(EXIT_FAILURE);
		}
		snprintf(value, 1024, "%lu", aval);
		ret = true;
	}
	if(ndtype == H5T_NATIVE_ULLONG) {
		unsigned long long aval;
		if(H5Aread(dattr, dtype, &aval) < 0) {
			std::cerr << "error: couldn't convert attribute " << name << "\n";
			exit(EXIT_FAILURE);
		}
		snprintf(value, 1024, "%lu", aval);
		ret = true;
	}
	if(ndtype == H5T_NATIVE_HBOOL) {
		bool aval;
		if(H5Aread(dattr, dtype, &aval) < 0) {
			std::cerr << "error: couldn't convert attribute " << name << "\n";
			exit(EXIT_FAILURE);
		}
		snprintf(value, 1024, "%d", aval);
		ret = true;
	}*/
	value[1023] = '\0';
	H5Tclose(dtype);
	H5Aclose(dattr);
	H5Gclose(dgroup);

	return true;
}

// create a data buffer
void *ECSB::createDataBuffer(unsigned long nmemb, hid_t dtype)
{
	void *buffer;

	if(!(buffer = calloc(nmemb, H5Tget_size(dtype)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	return buffer;
}

// free a data buffer
void ECSB::freeDataBuffer(void *buffer)
{
	free(buffer);
}

// create a data entry
void *ECSB::createDataEntry(hid_t dtype)
{
	void *entry;
	if(!(entry = malloc(H5Tget_size(dtype)))) {
		std::cerr << "error: couldn't allocate memory for data entry\n";
		exit(EXIT_FAILURE);
	}
	return entry;
}

// free a data entry
void ECSB::freeDataEntry(void *entry)
{
	free(entry);
}

// create a 3D array accessible by hdf5
unsigned long ***ECSB::create3DArrayULong(const hsize_t dims[3])
{
	unsigned long i, j, *buffer_tmp_tmp, **buffer_tmp, ***buffer;

	// allocate memory
	if(!(buffer_tmp_tmp = (unsigned long *) calloc(dims[0] * dims[1] * \
			dims[2], sizeof(unsigned long)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	if(!(buffer_tmp = (unsigned long **) calloc(dims[0] * dims[1], \
			sizeof(unsigned long *)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	if(!(buffer = (unsigned long ***) calloc(dims[0], \
			sizeof(unsigned long **)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}

	// fix pointers
	for(i = 0; i < dims[0]; i++) {
		buffer[i] = buffer_tmp + i * dims[1];
		for(j = 0; j < dims[1]; j++)
			buffer[i][j] = buffer_tmp_tmp + i * dims[1] * dims[2] + j * dims[2];
	}

	return buffer;
}

// create a 3D array accessible by hdf5
unsigned int ***ECSB::create3DArrayUInt(const hsize_t dims[3])
{
	unsigned int i, j, *buffer_tmp_tmp, **buffer_tmp, ***buffer;

	// allocate memory
	if(!(buffer_tmp_tmp = (unsigned int *) calloc(dims[0] * dims[1] * \
			dims[2], sizeof(unsigned int)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	if(!(buffer_tmp = (unsigned int **) calloc(dims[0] * dims[1], \
			sizeof(unsigned int *)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	if(!(buffer = (unsigned int ***) calloc(dims[0], \
			sizeof(unsigned int **)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}

	// fix pointers
	for(i = 0; i < dims[0]; i++) {
		buffer[i] = buffer_tmp + i * dims[1];
		for(j = 0; j < dims[1]; j++)
			buffer[i][j] = buffer_tmp_tmp + i * dims[1] * dims[2] + j * dims[2];
	}

	return buffer;
}

// free 3D array
void ECSB::free3DArrayULong(unsigned long ***buffer)
{
	free(buffer[0][0]);
	free(buffer[0]);
	free(buffer);
}

// free 3D array
void ECSB::free3DArrayUInt(unsigned int ***buffer)
{
	free(buffer[0][0]);
	free(buffer[0]);
	free(buffer);
}

// create a 2D array accessible by hdf5
double **ECSB::create2DArrayDouble(const hsize_t dims[2])
{
	unsigned long i;
	double **buffer;
	
	// allocate memory
	if(!(buffer = (double **) malloc(dims[0] * sizeof(double *)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	if(!(buffer[0] = (double *) malloc(dims[0] * dims[1] * \
			sizeof(double)))) {
		std::cerr << "error: couldn't allocate memory for buffer\n";
		exit(EXIT_FAILURE);
	}
	
	// fix pointers
	for(i = 1; i < dims[0]; i++)
		buffer[i] = buffer[0] + i * dims[1];

	return buffer;
}

// free 2D array
void ECSB::free2DArrayDouble(double **buffer)
{
	free(buffer[0]);
	free(buffer);
}
