// hdf5 dataset functions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#ifndef DATASET_H
#define DATASET_H

#include <hdf5.h>

namespace ECSB {

// byte datatype
#define DS_BYTE unsigned char

// get entry in two-dimensional array stored as one-dimensional array
#define get2DArrayEntry(base, size, i, j) (*((base) + ((size) * (i)) + (j)))

// get offset of compound datatype member by name
#define getEntryOffset(dtype, name) \
		(H5Tget_member_offset(dtype, H5Tget_member_index(dtype, name)))

// offsets for array entries
#define charOffset(a)	((a) * sizeof(char))
#define intOffset(a)	((a) * sizeof(int))
#define longOffset(a)	((a) * sizeof(long))
#define uLongOffset(a)	((a) * sizeof(unsigned long))
#define doubleOffset(a)	((a) * sizeof(double))

// get size of hdf5 datatype
hssize_t getDatasetSize(hid_t dset);

// create an hdf5 chunked property list
hid_t createChunkedPropertyList(const int rank, hsize_t *dims, \
		const hid_t dtype, unsigned long fill_value = 0);

// create an hdf5 file
hid_t createDatafile(const char *filename);

// open an hdf5 file
hid_t openDatafile(const char *filename);

// close an hdf5 datafile
void closeDatafile(hid_t dfile);

// create an hdf5 dataspace
hid_t createDataspace(const unsigned long size);
hid_t createDataspaceScaled(unsigned long *size);
hid_t createDataspace(int rank, const hsize_t *dims, \
		const hsize_t *maxdims = NULL);
hid_t createDataspaceScaled(int rank, hsize_t *dims, hsize_t *maxdims = NULL);

// clost an hdf5 dataspace
void closeDataspace(hid_t dspace);

// create an hdf5 dataset
hid_t createDataset(const char *name, hid_t dfile, hid_t dtype, hid_t dspace);

// open an hdf5 dataset
hid_t openDataset(const char *name, hid_t dfile);

// write an hdf5 dataset
void writeDataset(void *entry, void *buffer, hid_t dset, hid_t dtype, \
		hid_t dspace, unsigned long &buffer_index, \
		const unsigned long chunk_size);

// read all entries in an hdf5 dataset in one go
void readDataset(void *buffer, hid_t dset);

// read an hdf5 dataset in chunks
unsigned long readDataset(void *buffer, hid_t dset, hid_t dtype, \
		unsigned long &buffer_index, const unsigned long buffer_size);

// close an hdf5 dataset after reading
void closeDataset(hid_t dset);

// close an hdf5 dataset after writing
void closeDataset(hid_t dfile, hid_t dset, hid_t dtype, void *buffer, \
		unsigned long &buffer_index, const unsigned long chunk_size);

// create an hdf5 group
hid_t createGroup(const char *name, hid_t dfile);

// open an hdf5 group
hid_t openGroup(const char *name, hid_t dfile);

// close an hdf5 group
void closeGroup(hid_t group);

// write an attribute to an hdf5 datafile
void writeAttribute(const char *name, const char *value,hid_t dfile, \
		const char *group = "/");

// write an attribute to an hdf5 datafile
void writeAttribute(const char *name, const int value, hid_t dfile, \
		const char *group = "/");

// write an attribute to an hdf5 datafile
void writeAttribute(const char *name, const unsigned long value, hid_t dfile, \
		const char *group = "/");

// write an attribute to an hdf5 datafile
void writeAttribute(const char *name, const unsigned long long value, \
		hid_t dfile, const char *group = "/");

// write an attribute to an hdf5 datafile
void writeAttribute(const char *name, const double value, hid_t dfile, \
		const char *group = "/");

// write an attribute to an hdf5 datafile
void writeAttribute(const char *name, const bool value, hid_t dfile, \
		const char *group = "/");

// read an attribute from an hdf5 datafile
bool readAttribute(const char *name, char *&value, hid_t dfile, \
		const char *group = "/");

// read an attribute from an hdf5 datafile
bool readAttribute(const char *name, int &value, hid_t dfile, \
		const char *group = "/");

// read an attribute from an hdf5 datafile
bool readAttribute(const char *name, unsigned long &value, hid_t dfile, \
		const char *group = "/");

// read an attribute from an hdf5 datafile
bool readAttribute(const char *name, unsigned long long &value, hid_t dfile, \
		const char *group = "/");

// read an attribute from an hdf5 datafile
bool readAttribute(const char *name, double &value, hid_t dfile, \
		const char *group = "/");

// read an attribute from an hdf5 datafile
bool readAttribute(const char *name, bool &value, hid_t dfile, \
		const char *group = "/");

// convert an attribute from an hdf5 datafile to a string
bool attributeAsString(const char *name, char *&value, hid_t dfile, \
		const char *group = "/");

// create a data buffer
void *createDataBuffer(unsigned long nmemb, hid_t dtype);

// free a data buffer
void freeDataBuffer(void *buffer);

// create a data entry
void *createDataEntry(hid_t dtype);

// free a data entry
void freeDataEntry(void *entry);

// create a 3D array accessible by hdf5
unsigned long ***create3DArrayULong(const hsize_t dims[3]);

// create a 3D array accessible by hdf5
unsigned int ***create3DArrayUInt(const hsize_t dims[3]);

// free 3D array
void free3DArrayULong(unsigned long ***buffer);

// free 3D array
void free3DArrayUInt(unsigned int ***buffer);

// create a 2D array accessible by hdf5
double **create2DArrayDouble(const hsize_t dims[2]);

// free 2D array
void free2DArrayDouble(double **buffer);

}

#endif
