/*
 * File:   finitelattice.cpp
 * Author: Francisco J. Romero-Campero
 *
 * Created: January 2010
 */

#include "FiniteLattice.h"

using namespace ECSB;
using namespace std;
using namespace xmlpp;

//Default Constructor: It creates a lattice with a single point
finiteLattice::finiteLattice()
{
    name = "onePoint";
    dimension = 2;
    xmin = 0;
    xmax = 0;
    ymin = 0;
    ymax = 0;
}

//Copy constructor
finiteLattice::finiteLattice(const finiteLattice& orig)
{
  name = orig.name;
  dimension = orig.dimension;
  units = orig.units;
  scale = orig.scale;
  xmin = orig.xmin;
  xmax = orig.xmax;
  ymin = orig.ymin;
  ymax = orig.ymax;
  parameters = orig.parameters;
  basis = orig.basis;
  vertices = orig.vertices;
  neighbours = orig.neighbours;
}

//Default destructor
finiteLattice::~finiteLattice()
{

}

//Constructor from a file
finiteLattice::finiteLattice(std::string latticeFile)
{
    //Variable for the xml file with the definition of a lattice
    std::string xmlFile;

    //Getting the extesion of the file
    size_t pos = latticeFile.find('.');
    std::string fileExtension = latticeFile.substr(pos+1);

    //If the extension is lat we parse the file according to our BNF grammar
    //in order to generate the corresponding xml file.
    if(fileExtension == "lat")
    {
        //The lattice parser generated from our BNF grammar
        lattice_p latticeParser;

        //Generate a string with the content of the file
        std::string stringFile = fileToString(latticeFile);

        //We parse the content of the file to generate an abstract syntax tree
        tree_parse_info<> result = ast_parse(stringFile.c_str(),
                                             latticeParser,
                                             space_p);

        if (result.full) //When parsing is succesfull the xml file is generated
        {
            iter_t iter = result.trees.begin();
            xmlFile = generateLatticeXML(iter);
        }
        else //Failed parsing
        {
            std::map<unsigned long, std::string> fileLines = fileToLines(latticeFile);
            unsigned long lineNumber = findLine(result.stop,fileLines);

            cerr << "Parsing of file " << latticeFile;
            cerr << " according to BNF failed in line: " << lineNumber << std::endl;
            //std::cout << result.stop << std::endl;
            exit(EXIT_FAILURE);
        }
    }
    //Specification in xml format does not require any convertion
    else if(fileExtension == "xml")
    {
        xmlFile = latticeFile;
    }

    //Finally we create the lattice using the xml file.
    createFromXML(xmlFile);
}

//Function to create a finite lattice object from an XML file
void finiteLattice::createFromXML(std::string xmlFile)
{
  //Parsing using the DOM parser
  DomParser parser;
  parser.parse_file(xmlFile);
  if(!parser)
  {
      cerr << "Parsing unsuccesfull for file " << xmlFile << std::endl;
      exit(EXIT_FAILURE);
  }

  //Root node of the file
  xmlpp::Node* latticeDeclaration = parser.get_document()->get_root_node();

  //Instantiation of the different attributes of the finite lattice

  //Name, dimension, units, scales and bounds
  xmlpp::Element* latticeElement = dynamic_cast<xmlpp::Element*>(latticeDeclaration);
  this->name = (latticeElement->get_attribute("name"))->get_value();
  this->dimension = fromStringTo<int>((latticeElement->get_attribute("dimension"))->get_value());

  xmlpp::Attribute* unitsDeclaration = latticeElement->get_attribute("units");
  if(unitsDeclaration != NULL)
  {
    this->units = unitsDeclaration->get_value();
  }

  xmlpp::Attribute* scaleDeclaration = latticeElement->get_attribute("scale");
  if(scaleDeclaration != NULL)
  {
      this->scale = fromStringTo<int>(scaleDeclaration->get_value());
  }

  xmlpp::Attribute* xminDeclaration =latticeElement->get_attribute("xmin");
  this->xmin = fromStringTo<int>(xminDeclaration->get_value());

  xmlpp::Attribute* xmaxDeclaration =latticeElement->get_attribute("xmax");
  this->xmax = fromStringTo<int>(xmaxDeclaration->get_value());

  xmlpp::Attribute* yminDeclaration =latticeElement->get_attribute("ymin");
  this->ymin = fromStringTo<int>(yminDeclaration->get_value());

  xmlpp::Attribute* ymaxDeclaration =latticeElement->get_attribute("ymax");
  this->ymax = fromStringTo<int>(ymaxDeclaration->get_value());

  xmlpp::Node::NodeList parametersDeclaration = latticeDeclaration->get_children("parameter");
  for(xmlpp::Node::NodeList::iterator par = parametersDeclaration.begin();
      par != parametersDeclaration.end();
      ++par)
    {
      xmlpp::Element* parElement = dynamic_cast<xmlpp::Element*>(*par);
      std::string parName = (parElement->get_attribute("name"))->get_value();
      double parValue = fromStringTo<double>((parElement->get_attribute("value"))->get_value());
      this->parameters.insert(pair<std::string,double>(parName,parValue));
    }

  //Basis declaration
  xmlpp::Node* basisDeclaration = (latticeDeclaration->get_children("basis")).front();
  xmlpp::Node::NodeList vectors = basisDeclaration->get_children("vector");

  for(xmlpp::Node::NodeList::iterator vect = vectors.begin();
      vect != vectors.end();
      ++vect)
    {
      xmlpp::Element* vectorElement = dynamic_cast<xmlpp::Element*>(*vect);

      xmlpp::Attribute* xDim = vectorElement->get_attribute("x");
      xmlpp::Attribute* yDim = vectorElement->get_attribute("y");

      std::vector<std::string> basisVector;
      if(xDim != NULL)
	basisVector.push_back(xDim->get_value());
      if(yDim != NULL)
	basisVector.push_back(yDim->get_value());

      (this->basis).push_back(basisVector);
    }

  //Neighbours declaration
  xmlpp::Node* neighboursDeclaration = (latticeDeclaration->get_children("neighbours")).front();
  xmlpp::Node::NodeList neighVectors = neighboursDeclaration->get_children("vector");

  for(xmlpp::Node::NodeList::iterator vect = neighVectors.begin();
      vect != neighVectors.end();
      ++vect)
    {
      xmlpp::Element* vectorElement = dynamic_cast<xmlpp::Element*>(*vect);

      xmlpp::Attribute* xDim = vectorElement->get_attribute("x");
      xmlpp::Attribute* yDim = vectorElement->get_attribute("y");

      std::vector<std::string> neighVector;
      if(xDim != NULL)
	neighVector.push_back(xDim->get_value());
      if(yDim != NULL)
	neighVector.push_back(yDim->get_value());

      (this->neighbours).push_back(neighVector);
    }

  //Vertices declaration
  xmlpp::Node::NodeList verticesDeclaration = latticeDeclaration->get_children("vertices");

  if(!verticesDeclaration.empty())
  {
      xmlpp::Node* vertices = verticesDeclaration.front();
      xmlpp::Node::NodeList verticesVectors = vertices->get_children("vector");


      for(xmlpp::Node::NodeList::iterator vect = verticesVectors.begin();
        vect != verticesVectors.end();
        ++vect)
        {
            xmlpp::Element* vectorElement = dynamic_cast<xmlpp::Element*>(*vect);

            xmlpp::Attribute* xDim = vectorElement->get_attribute("x");
            xmlpp::Attribute* yDim = vectorElement->get_attribute("y");

            std::vector<std::string> vertexVector;
            if(xDim != NULL)
                vertexVector.push_back(xDim->get_value());
            if(yDim != NULL)
                vertexVector.push_back(yDim->get_value());

            (this->vertices).push_back(vertexVector);
        }
  }

}


void finiteLattice::show()
{
  std::cout << "------------------------------" << std::endl;
  std::cout << "Lattice: " << this->name << std::endl;
  std::cout << "\tdimension:" << this->dimension << std::endl;
  std::cout << "\tunits: " << this->units << std::endl;
  std::cout << "\tscale: " << this->scale << std::endl;
  std::cout << "\txmin: " << this->xmax << std::endl;
  std::cout << "\txmax: " << this->xmax << std::endl;

  if(this->dimension >= 2)
  {
      std::cout << "\tymin: " << this->ymin << std::endl;
      std::cout << "\tymax: " << this->ymax << std::endl;
  }

  std::cout << "\tparameters:" << std::endl;
  for(std::map<std::string,double>::iterator par = (this->parameters).begin();
      par != (this->parameters).end();
      ++par)
    {
      std::cout << "\t\tparameter: " << par->first << " = " << par->second;
      std::cout << std::endl;
    }

  std::cout << "\tbasis:" << std::endl;
  for(std::vector< std::vector<string> >::iterator vect = (this->basis).begin();
      vect != (this->basis).end();
      ++vect)
    {
      std::cout << "\t\tvector: ";
      for(std::vector<std::string>::iterator c = vect->begin();
	  c != vect->end();
	  ++c)
	{
	  std::cout << *c << " ";
	}
      std::cout << std::endl;

    }

  std::cout << std::endl <<"\tneighbours:" << std::endl;
  for(std::vector< std::vector<string> >::iterator vect = (this->neighbours).begin();
      vect != (this->neighbours).end();
      ++vect)
    {
      std::cout << "\t\tvector: ";
      for(std::vector<std::string>::iterator c = vect->begin();
	  c != vect->end();
	  ++c)
	{
	  std::cout << *c << " ";
	}
      std::cout << std::endl;

    }

  std::cout << std::endl <<"\tvertices:" << std::endl;
  for(std::vector< std::vector<string> >::iterator vect = (this->vertices).begin();
      vect != (this->vertices).end();
      ++vect)
    {
      std::cout << "\t\tvector: ";
      for(std::vector<std::string>::iterator c = vect->begin();
	  c != vect->end();
	  ++c)
	{
	  std::cout << *c << " ";
	}
      std::cout << std::endl;

    }

  std::cout << "------------------------------" << std::endl;

}
