/*
 * File:   LPPRule.cpp
 * Author: Francisco J. Romero-Campero
 *
 * Created on 18 January 2010, 12:16
 */

#include <iostream>

#include "LPPRule.h"
#include "ModuleLibrary.h"
#include "Misc.h"

using namespace std;
using namespace xmlpp;
using namespace ECSB;

//Default constructor
LPPRule::LPPRule()
{

}

//Copy constructor
LPPRule::LPPRule(const LPPRule& orig)
{
    this->constant = orig.constant;
    this->label = orig.label;
    this->labelLhs = orig.labelLhs;
    this->lhsInside = orig.lhsInside;
    this->lhsOutside = orig.lhsOutside;
    this->lhsTranslocation = orig.lhsTranslocation;
    this->name = orig.name;
    this->rhsInside = orig.rhsInside;
    this->rhsOutside = orig.rhsOutside;
    this->rhsTranslocation = orig.rhsTranslocation;
    this->ruleVector = orig.ruleVector;
    this->stringConstant = orig.stringConstant;
    this->type = orig.type;
}

//Default destructor
LPPRule::~LPPRule()
{

}

//Constructor from XML specification
LPPRule::LPPRule(xmlpp::Node* ruleDeclaration)
{
    //We cast rule_node to Element to get its attributes
    xmlpp::Element* rule = dynamic_cast<xmlpp::Element*>(ruleDeclaration);

    //Getting the name of the rule
    this->name = (rule->get_attribute("name"))->get_value();
    //Getting the type of the rule
    xmlpp::Attribute* typeDeclaration = rule->get_attribute("type");
    if(typeDeclaration != NULL)
    {
        this->type = typeDeclaration->get_value();
    }

    //Getting the vector associated with the rule if it is a translocation rule
    xmlpp::Attribute* vectorDeclaration = rule->get_attribute("vector");
    if(vectorDeclaration != NULL)
    {
        std::string vectorString = vectorDeclaration->get_value();
        std::string::size_type commaPosition = vectorString.find(",", 0);
        std::string xString = vectorString.substr(0,commaPosition);
        std::string yString = vectorString.substr(commaPosition+1);
        this->ruleVector.push_back(fromStringTo<int>(xString));
        this->ruleVector.push_back(fromStringTo<int>(yString));
        this->ruleVector.push_back(0);
    }

    //Getting the constant associated with the rule distinguishing between a
    //rule defined in a module without constant value and a rule with a constant
    //value associated
    xmlpp::Attribute* ruleConstant = rule->get_attribute("constant");
    if(ruleConstant == NULL)
    {
        cerr << "Error: rule " << this->name << " without constant";
        cerr << std::endl;
        exit(EXIT_FAILURE);
    }
    else
    {
        this->stringConstant = ruleConstant->get_value();
    }

    xmlpp::Attribute* constantValue = rule->get_attribute("value");
    if(constantValue == NULL)
    {
        this->constant = 0;
    }
    else
    {
        this->constant = fromStringTo<double>(constantValue->get_value());
    }

    //Getting the LHS
    xmlpp::Node* lhs = (ruleDeclaration->get_children("lhs")).front();

    //Getting the outside if it is not empty
    xmlpp::Node::NodeList listOfOutsideObjects =
            lhs->get_children("listOfOutsideObjects");
    if(!listOfOutsideObjects.empty())
    {
        xmlpp::Node* nodeOutsideObjects = listOfOutsideObjects.front();
        xmlpp::Node::NodeList outsideObjects =
                nodeOutsideObjects->get_children("object");
        for(xmlpp::Node::NodeList::iterator obj = outsideObjects.begin();
	    obj != outsideObjects.end();
	    ++obj)
        {
            xmlpp::Element* object = dynamic_cast<xmlpp::Element*>(*obj);
            std::string objectName =
                    (object->get_attribute("name"))->get_value();
            this->lhsOutside.push_back(objectName);
        }
    }

    //Getting the inside if it is not empty
    xmlpp::Node::NodeList listOfInsideObjects =
            lhs->get_children("listOfInsideObjects");
    if(!listOfInsideObjects.empty())
    {
        xmlpp::Node* nodeInsideObjects = listOfInsideObjects.front();
        xmlpp::Node::NodeList insideObjects =
                nodeInsideObjects->get_children("object");
        for(xmlpp::Node::NodeList::iterator obj = insideObjects.begin();
	    obj != insideObjects.end();
	    ++obj)
        {
            xmlpp::Element* object = dynamic_cast<xmlpp::Element*>(*obj);
            std::string objectName =
                    (object->get_attribute("name"))->get_value();
            this->lhsInside.push_back(objectName);
        }

        //Get label of the lhs inside objects
        xmlpp::Element* objectsInsideElement =
                dynamic_cast<xmlpp::Element*>(nodeInsideObjects);
        this->label =
                (objectsInsideElement->get_attribute("label"))->get_value();
    }

    //Getting the translocated objects if not empty
    xmlpp::Node::NodeList listOfTranslocatedObjects =
            lhs->get_children("listOfTranslocatedObjects");
    if(!listOfTranslocatedObjects.empty())
    {
        xmlpp::Node* nodeTranslocatedObjects =
                listOfTranslocatedObjects.front();
        xmlpp::Node::NodeList translocatedObjects =
                nodeTranslocatedObjects->get_children("object");
        for(xmlpp::Node::NodeList::iterator obj = translocatedObjects.begin();
	    obj != translocatedObjects.end();
	    ++obj)
	{
            xmlpp::Element* object = dynamic_cast<xmlpp::Element*>(*obj);
	    std::string objectName =
                    (object->get_attribute("name"))->get_value();
	    lhsTranslocation.push_back(objectName);
	}

        //Get label of the lhs translocated objects
        xmlpp::Element* translocatedElement =
                dynamic_cast<xmlpp::Element*>(nodeTranslocatedObjects);
        this->labelLhs =
                (translocatedElement->get_attribute("label"))->get_value();
    }

    //Getting the RHS
    xmlpp::Node* rhs = (ruleDeclaration->get_children("rhs")).front();

    //Getting the outside if it is not empty
    listOfOutsideObjects = rhs->get_children("listOfOutsideObjects");
    if(!listOfOutsideObjects.empty())
    {
        xmlpp::Node* nodeOutsideObjects = listOfOutsideObjects.front();
        xmlpp::Node::NodeList outsideObjects =
                nodeOutsideObjects->get_children("object");
        for(xmlpp::Node::NodeList::iterator obj = outsideObjects.begin();
	    obj != outsideObjects.end();
	    ++obj)
        {
            xmlpp::Element* object = dynamic_cast<xmlpp::Element*>(*obj);
            std::string objectName =
                    (object->get_attribute("name"))->get_value();
            this->rhsOutside.push_back(objectName);
        }
    }

    //Getting the inside if it is not empty
    listOfInsideObjects = rhs->get_children("listOfInsideObjects");
    if(!listOfInsideObjects.empty())
    {
        xmlpp::Node* nodeInsideObjects = listOfInsideObjects.front();
        xmlpp::Node::NodeList insideObjects =
                nodeInsideObjects->get_children("object");
        for(xmlpp::Node::NodeList::iterator obj = insideObjects.begin();
	    obj != insideObjects.end();
	    ++obj)
        {
            xmlpp::Element* object = dynamic_cast<xmlpp::Element*>(*obj);
            std::string objectName =
                    (object->get_attribute("name"))->get_value();
            this->rhsInside.push_back(objectName);
        }

        //Get label of the lhs inside objects
        xmlpp::Element* objectsInsideElement =
                dynamic_cast<xmlpp::Element*>(nodeInsideObjects);
        this->label =
                (objectsInsideElement->get_attribute("label"))->get_value();
    }

    //Getting the translocated objects if not empty
    listOfTranslocatedObjects = rhs->get_children("listOfTranslocatedObjects");
    if(!listOfTranslocatedObjects.empty())
    {
        xmlpp::Node* nodeTranslocatedObjects =
                listOfTranslocatedObjects.front();
        xmlpp::Node::NodeList translocatedObjects =
                nodeTranslocatedObjects->get_children("object");
        for(xmlpp::Node::NodeList::iterator obj = translocatedObjects.begin();
	    obj != translocatedObjects.end();
	    ++obj)
	{
            xmlpp::Element* object = dynamic_cast<xmlpp::Element*>(*obj);
	    std::string objectName =
                    (object->get_attribute("name"))->get_value();
	    rhsTranslocation.push_back(objectName);
	}
        //Get label of the lhs translocated objects
        xmlpp::Element* translocatedElement =
                dynamic_cast<xmlpp::Element*>(nodeTranslocatedObjects);
        this->label =
                (translocatedElement->get_attribute("label"))->get_value();
    }

    //Determine the type of rule depending on the objects in the rule
    if(!lhsTranslocation.empty())
    {
        this->type = "translocation";
    }
    else
    {
        this->type = "boundary";
    }
}


std::list<std::string> LPPRule::getObjects()
{
  std::list<std::string> listLhsOutside(lhsOutside.begin(),lhsOutside.end());
  std::list<std::string> listLhsInside(lhsInside.begin(),lhsInside.end());
  std::list<std::string> listLhsTranslocation(lhsTranslocation.begin(),
                                              lhsTranslocation.end());
  std::list<std::string> listRhsOutside(rhsOutside.begin(),rhsOutside.end());
  std::list<std::string> listRhsInside(rhsInside.begin(),rhsInside.end());
  std::list<std::string> listRhsTranslocation(rhsTranslocation.begin(),
                                              rhsTranslocation.end());

  std::list<std::string> objects = listRhsInside;

  std::list<std::string>::iterator it = objects.begin();

  objects.splice(it,listRhsOutside);
  objects.splice(it,listRhsTranslocation);
  objects.splice(it,listLhsInside);
  objects.splice(it,listLhsOutside);
  objects.splice(it,listLhsTranslocation);
  objects.sort();
  objects.unique();

  return objects;
}

//Auxiliary function to substitute objects in a multiset
std::vector<std::string> LPPRule::substituteMultiset(
                                map<std::string,std::string> substitution,
                                std::vector<std::string> objects)
{
    for(map<std::string,std::string>::iterator it = substitution.begin();
	it != substitution.end();
	++it)
    {
        for(std::vector<std::string>::iterator o = objects.begin();
	    o != objects.end();
	    ++o)
        {
            size_t pos = o->find(it->first);
	    while(pos != o->npos)
            {
                o->replace(pos,(it->first).length(),it->second);
		pos = o->find(it->first,pos+1);
            }
        }
    }

    return objects;
}

//Function to substitute the objects in a rule according to a substitution map
void LPPRule::substituteObjects(map<std::string,std::string> substitution)
{
    lhsOutside = substituteMultiset(substitution, lhsOutside);
    lhsInside = substituteMultiset(substitution, lhsInside);
    lhsTranslocation = substituteMultiset(substitution, lhsTranslocation);
    rhsOutside = substituteMultiset(substitution, rhsOutside);
    rhsInside = substituteMultiset(substitution, rhsInside);
    rhsTranslocation = substituteMultiset(substitution, rhsTranslocation);
}

//Function to substitute the constants of a rule according to a substitution map
void LPPRule::substituteConstants(map<std::string,std::string> substitution)
{
    if(!substitution.empty())
    {
        if(this->constant == 0)
        {
            if(isNumeric(substitution[this->stringConstant]))
            {
                this->constant = fromStringTo<double>(substitution[this->stringConstant]);
            }
            else
            {
                this->stringConstant = substitution[this->stringConstant];
            }
        }
    }
}

//Function to substitute the constants of a rule according to a substitution map
void LPPRule::substituteConstants(map<std::string,double> substitution)
{
    if(!substitution.empty())
    {
        if(this->constant == 0)
        {
            this->constant = substitution[this->stringConstant];
        }
        else
        {
            this->constant = substitution[this->stringConstant];
            std::cout << "Warning: the constant " << this->stringConstant;
            std::cout << " already had a value associated that has been ";
            std::cout << " overwritten" << std::endl;
        }
    }
}

//Function to substitute the labels of a rule according to a substitution map
void LPPRule::substituteLabels(map<std::string,std::string> substitution)
{
    if(!substitution.empty())
    {
      this->labelLhs = substitution[this->labelLhs];
      this->label = substitution[this->label];
    }
}

//Displayer
void LPPRule::show()
{
  if(this->type == "boundary")
    {
      // We show the objects outside on the left hand side
      for(std::vector<std::string>::iterator obj = lhsOutside.begin();
	  obj != lhsOutside.end();
	  ++obj)
	{
	  std::cout << *obj << " " ;
	}

      std::cout << "[ ";

      // We show the objects inside on the left hand side
      for(std::vector<std::string>::iterator obj = lhsInside.begin();
	  obj != lhsInside.end();
	  ++obj)
	{
	  std::cout << *obj << " " ;
	}

      std::cout << "]_" << label << " --> ";

      // We show the objects outside on the right hand side
      for(std::vector<std::string>::iterator obj = rhsOutside.begin();
	  obj != rhsOutside.end();
	  ++obj)
	{
	  std::cout << *obj << " " ;
	}

      std::cout << "[ ";

      // We show the objects inside on the right hand side
      for(std::vector<std::string>::iterator obj = rhsInside.begin();
	  obj != rhsInside.end();
	  ++obj)
	{
	  std::cout << *obj << " " ;
	}

      std::cout << "]_" << label << "        ";
    }
  else if(this->type == "translocation")
    {
      //We show the objects on the left hand side
      std::cout << "[ " ;

      for(std::vector<std::string>::iterator obj = lhsTranslocation.begin();
	  obj != lhsTranslocation.end();
	  ++obj)
	{
	  std::cout << *obj << " ";
	}
      std::cout << "]_" << this->labelLhs << "=(" << this->ruleVector[0];
      std::cout << "," << this->ruleVector[1] << "," << this->ruleVector[2];
      std::cout << ")=";
      std::cout << "[ ] --> [ ]_" << labelLhs << "=(" << this->ruleVector[0];
      std::cout << "," << this->ruleVector[1] << "," << this->ruleVector[2];
      std::cout << ")=";
      std::cout << "[ ";

      for(std::vector<std::string>::iterator obj = rhsTranslocation.begin();
	  obj != rhsTranslocation.end();
	  ++obj)
	{
	  std::cout << *obj << " ";
	}

      std::cout << "]" << "\t";
    }


  std::cout << this->stringConstant << " = " << this->constant << std::endl;
}

///////////////////////////////////////
// miscellaneos function definitions //
///////////////////////////////////////

std::vector<LPPRule> ECSB::expand(xmlpp::Node *moduleDeclaration)
{
    std::map<std::string,ModuleDefinition> partialLibrary;
    std::vector<LPPRule> rules = expand(moduleDeclaration,partialLibrary);
    return rules;
}


std::vector<LPPRule> ECSB::expand(xmlpp::Node *moduleDeclaration,
        std::map<std::string,ModuleDefinition> partialLibrary)
{
  //Information about the module and the library
  xmlpp::Element* module = dynamic_cast<xmlpp::Element*>(moduleDeclaration);
  std::string moduleName = (module->get_attribute("name"))->get_value();
  std::string file = (module->get_attribute("file"))->get_value();

  //Obtain the module definition from the corresponding library or
  //partial library
  map<std::string,ModuleDefinition> modules;
  if(file == "this")
  {
      modules = partialLibrary;
  }
  else
  {
      ModuleLibrary moduleLibrary(file);
      modules = moduleLibrary.getModules();
  }

  if(modules.find(moduleName) == modules.end())
  {
      cerr << "Error: no module " << moduleName << " in library in file ";
      cerr << file << std::endl;
      exit(EXIT_FAILURE);
  }

  ModuleDefinition moduleDefinition = modules[moduleName];
  moduleDefinition = renameVariables(moduleDefinition);



  //Create substitutions vectors
  xmlpp::Attribute* objectsAttribute = module->get_attribute("objects");
  if(objectsAttribute == NULL)
  {
      cerr << "Error: no objects substitution for module " << moduleName;
      exit(EXIT_FAILURE);
  }
  std::string objectList = objectsAttribute->get_value();
  map<std::string,std::string> objectSubstitution =
    createSubstitution(moduleDefinition.getObjects(),
		       convertToVector(objectList));

  xmlpp::Attribute* constantsAttribute = module->get_attribute("constants");
  if(objectsAttribute == NULL)
  {
      cerr << "Error: no constants substitution for module " << moduleName << std::endl;
      exit(EXIT_FAILURE);
  }

  std::string constantList = constantsAttribute->get_value();
  map<std::string,std::string> constantSubstitution =
    createSubstitution(moduleDefinition.getConstants(),
		       convertToVector(constantList));

  xmlpp::Attribute* labelsAttribute = module->get_attribute("labels");
  if(labelsAttribute == NULL)
  {
      cerr << "Error: no labels substitution for module " << moduleName << std::endl;
      exit(EXIT_FAILURE);
  }

  std::string labelList = labelsAttribute->get_value();
  map<std::string,std::string> labelSubstitution =
    createSubstitution(moduleDefinition.getLabels(),
		       convertToVector(labelList));

  std::vector<LPPRule> finalRules = moduleDefinition.instantiate(objectSubstitution, \
  		constantSubstitution, labelSubstitution);

  return finalRules;

}

//Auxiliar remove duplicate rules
std::vector<LPPRule> ECSB::removeDuplicates(std::vector<LPPRule> rules)
{
  std::vector<LPPRule> uniqueRules;
  int pos = 0;
  for(std::vector<LPPRule>::iterator r = rules.begin(); r != rules.end(); ++r)
    {
      pos++;
      if(ruleIn(*r,rules,pos)==0)
	uniqueRules.push_back(*r);
    }
 return uniqueRules;
}

//Auxiliar to associate unique names to rules after expansion and instantiation
std::vector<LPPRule> ECSB::individualRuleNames(std::vector<LPPRule> rules, std::string lab)
{
  unsigned long index = 1;

  for(std::vector<LPPRule>::iterator r = rules.begin(); r != rules.end(); ++r)
    {
      std::string ruleName = "r_" + lab + "_" + somethingToString<double>(index);
      r->setName(ruleName);
      index++;
    }
  return rules;
}

//Auxiliar function to determine whether or not a LPPRule is in a given vector of rules
//starting from a specific position in the vector specified by an iterator
int ECSB::ruleIn(LPPRule rule, std::vector<LPPRule> rules, int pos)
{
  int h = 0;
  for(std::vector<LPPRule>::iterator r = rules.begin(); r != rules.end(); ++r)
    {
      if( h >= pos )
	{
	  if(ruleEqual(rule,*r)==1)
	    return 1;
	}
      else
	{
	  h++;
	}
    }
  return 0;
}

int ECSB::ruleEqual(LPPRule r1, LPPRule r2)
{
  if(r1.getType() == r2.getType() &&
     r1.getConstant() == r2.getConstant() &&
     r1.getLabelLhs() == r2.getLabelLhs() &&
     r1.getLabel() == r2.getLabel() &&
     r1.getVector() == r2.getVector() &&
     equalityVector(r1.getLhsOutside(), r2.getLhsOutside()) &&
     equalityVector(r1.getLhsInside(), r2.getLhsInside()) &&
     equalityVector(r1.getLhsTranslocation(), r2.getLhsTranslocation()) &&
     equalityVector(r1.getRhsOutside(), r2.getRhsOutside()) &&
     equalityVector(r1.getRhsInside(), r2.getRhsInside()) &&
     equalityVector(r1.getRhsTranslocation(), r2.getRhsTranslocation())
     )
    {
      return 1;
    }
  else
    {
      return 0;
    }
}
