/*
 * File:   LPPsystem.cpp
 * Author: fran
 *
 * Created on 14 January 2010, 14:31
 */

#include <vector>


#include "LPPsystem.h"
#include "FiniteLattice.h"

using namespace std;
using namespace ECSB;
using namespace xmlpp;

//Default constructor
LPPsystem::LPPsystem()
{

}

//Copy constructor
LPPsystem::LPPsystem(const LPPsystem& orig)
{
  name = orig.name;
  pSystems = orig.pSystems;
  lattice = orig.lattice;
  positions = orig.positions;
}

//Default destructor
LPPsystem::~LPPsystem()
{

}

LPPsystem::LPPsystem(std::string LPPsystemFile)
{
    //Variable for the xml file with the definition of a lattice
    std::string xmlFile;

    //Getting the extesion of the file
    size_t pos = LPPsystemFile.find('.');
    std::string fileExtension = LPPsystemFile.substr(pos+1);

    //If the extension is lpp we parse the file according to our BNF grammar
    //in order to generate the corresponding xml file.
    if(fileExtension == "lpp" || fileExtension == "sbml")
    {
        if(fileExtension == "sbml")
        {
            LPPsystemFile = generateModel(LPPsystemFile.c_str());
        }

        //The LPP system parser generated from our BNF grammar
        LPPsystem_p lppParser;    //  Our parser

        //Generate a string with the content of the file
        std::string stringFile = fileToString(LPPsystemFile);

        //We parse the content of the file to generate an abstract syntax tree
        tree_parse_info<> result = ast_parse(stringFile.c_str(),
                                             lppParser,
                                             space_p);

        if (result.full) //When parsing is succesfull the xml file is generated
        {
            iter_t iter = result.trees.begin();
            xmlFile = generateLPPXML(iter);
        }
        else //Failed parsing
        {
            std::map<unsigned long, std::string> fileLines = fileToLines(LPPsystemFile);
            unsigned long lineNumber = findLine(result.stop,fileLines);

            cerr << "Parsing of file " << LPPsystemFile;
            cerr << " according to BNF failed in line: " << lineNumber << std::endl;
            //std::cout << result.stop << std::endl;
            exit(EXIT_FAILURE);
        }
    }
    //Specification in xml format
    else if(fileExtension == "xml")
    {
        xmlFile = LPPsystemFile;
    }
    else
    {
        cerr << "Format not supported " << LPPsystemFile << std::endl;
        exit(EXIT_FAILURE);
    }

    //Finally we create the LPP system using the xml file.
    createFromXML(xmlFile);

}

void LPPsystem::createFromXML(std::string xmlFile)
{
  //Parsing using the DOM parser
  DomParser parser;
  parser.parse_file(xmlFile);
  if(!parser)
    {
      cerr << "Parsing unsuccesfull for file " << xmlFile << std::endl;
      exit(EXIT_FAILURE);
    }

  //Get the root node of the LPP system
  xmlpp::Node* modelDeclaration = parser.get_document()->get_root_node();

  //We get the name of the model
  xmlpp::Element* model = dynamic_cast<xmlpp::Element*>(modelDeclaration);
  if(model->get_attribute("name") == NULL)
  {
      cerr << "Model without name" << std::endl;
      exit(EXIT_FAILURE);
  }
  else
      this->name = (model->get_attribute("name"))->get_value();

  //Get the declaration of the different components of the model
  xmlpp::Node::NodeList components = modelDeclaration->get_children();
  xmlpp::Node::NodeList pSystemsDeclaration;
  xmlpp::Node* latticeDeclaration = NULL;
  xmlpp::Node* spatialDistributionDeclaration = NULL;

  for(xmlpp::Node::NodeList::iterator comp = components.begin();
      comp != components.end();
      ++comp)
  {
      if((*comp)->get_name() == "listOfPSystems")
	{
	  pSystemsDeclaration = (*comp)->get_children("PSystem");
	}
      else if ((*comp)->get_name() == "lattice")
	{
	  latticeDeclaration = *comp;
	}
      else if ((*comp)->get_name() == "spatialDistribution")
	{
	  spatialDistributionDeclaration = *comp;
	}
  }

  // sanity check
  if(pSystemsDeclaration.empty())
  {
      cerr << "error: no P system definition\n";
      exit(EXIT_FAILURE);
  }
  if(!latticeDeclaration)
  {
      cerr << "error: no lattice definition\n";
      exit(EXIT_FAILURE);
  }
  if(!spatialDistributionDeclaration)
  {
      cerr << "error: no cellular automata definition\n";
      exit(EXIT_FAILURE);
  }

  //Construct the different components from their declarations
  //Construct the list of P systems
  addPSystems(pSystemsDeclaration);

  //Construct the lattice
  xmlpp::Element* latticeElement = dynamic_cast<xmlpp::Element*>(latticeDeclaration);
  xmlpp::Attribute* latticeFile = latticeElement->get_attribute("file");

  if(latticeFile != NULL)
  {
      std::string latticeFileName = latticeFile->get_value();
      finiteLattice lat(latticeFileName);
      this->lattice = lat;

  }
  else
  {
      cerr << "error: the lattice specification must be in a separte file\n";
      exit(EXIT_FAILURE);
  }

  //Construct the spatialDistribution
  spatialDistribution positionedCells(spatialDistributionDeclaration,this->pSystems);

  //Mutate the cellular automata according to the noise if specified
  double noise = positionedCells.getNoise();

  if (noise > 0)
  {
    int numberMutations = positionedCells.getNumberMutations();

    for(int i = 0; i < numberMutations; ++i)
        positionedCells.addNoise(noise,
                                (this->lattice).getXMin(),
                                (this->lattice).getXMax(),
                                (this->lattice).getYMin(),
                                (this->lattice).getYMax());
  }

  this->positions = positionedCells;

}

//Add the list of P systems to the LPP system
void LPPsystem::addPSystems(xmlpp::Node::NodeList pSystemsDeclaration)
{

   for(xmlpp::Node::NodeList::iterator pSyst = pSystemsDeclaration.begin();
       pSyst != pSystemsDeclaration.end();
       ++pSyst)
   {
       //Name of the PSystem and file where it is specified
       xmlpp::Element* pSystemElement = dynamic_cast<xmlpp::Element*>(*pSyst);

       xmlpp::Attribute* nameDeclaration = pSystemElement->get_attribute("name");
       std::string pSystemName = nameDeclaration->get_value();

       xmlpp::Attribute* fileDeclaration = pSystemElement->get_attribute("file");
       std::string fileName = fileDeclaration->get_value();

       //Constructing the P system using the file with its specification
       PSystem pSystem(fileName);

       //Changing the name of the P system according to the specification in
       //the LPP system
       pSystem.setName(pSystemName);

       //Inserting the created P system
       this->pSystems[pSystem.getName()] = pSystem;
   }
}

PSystem LPPsystem::getPSystem(std::string pSystemName)
{
  return (this->pSystems)[pSystemName];
}

std::vector<std::string> LPPsystem::getCombinedAlphabets()
{
  //Auxiliary list
  std::list<std::string> temporalList;

  //Accumulating the alphabets from each P system into the auxiliary list
  for(std::map<std::string,PSystem>::iterator pSyst = this->pSystems.begin();
      pSyst != this->pSystems.end();
      ++pSyst)
    {
      std::vector<std::string> pSystemAlphabet = (pSyst->second).getAlphabet();

      //Adding the objects from each particular alphabet
      for(std::vector<std::string>::iterator o = pSystemAlphabet.begin();
	  o != pSystemAlphabet.end();
	  ++o)
	{
	  temporalList.push_back(*o);
	}
    }

  //Eliminate duplicates
  temporalList.sort();
  temporalList.unique();

  //Convert into vector
  std::vector<std::string> combinedAlphabets(temporalList.begin(),temporalList.end());

  return combinedAlphabets;
}

std::map<std::string,LPPRuleSet> LPPsystem::getCombinedRuleSets()
{

  //Variable storing the result
  std::map<std::string,ECSB::LPPRuleSet> combinedRuleSets;

  for(std::map<std::string,PSystem>::iterator pSyst = (this->pSystems).begin();
      pSyst != (this->pSystems).end();
      ++pSyst)
    {
      std::string firstLabel = pSyst->first;

      //Adding the rules from each particular P system
      std::map<std::string,ECSB::LPPRuleSet> ruleSets = (pSyst->second).getRuleSets();
      for(std::map<std::string,ECSB::LPPRuleSet>::iterator rSet = ruleSets.begin();
	  rSet != ruleSets.end();
	  ++rSet)
	{
	  //The global label of the rule set is the combination of the P system
          //name and its label in the particular P system
	  std::string secondLabel = rSet->first;
	  std::string labelRuleSet = firstLabel + secondLabel;

          combinedRuleSets[labelRuleSet] = rSet->second;
	}
    }
  return combinedRuleSets;
}

std::map<std::string,ECSB::InitialMultiset> LPPsystem::getCombinedInitialMultisets()
{

  std::map<std::string,InitialMultiset> combinedInitialMultisets;

  for(std::map<std::string,PSystem>::iterator pSyst = (this->pSystems).begin();
      pSyst != (this->pSystems).end();
      ++pSyst)
    {
      std::string firstLabel = pSyst->first;
      std::map<std::string,ECSB::InitialMultiset> inis = (pSyst->second).getInitialMultisets();
      for(std::map<std::string,ECSB::InitialMultiset>::iterator ini = inis.begin();
	  ini != inis.end();
	  ++ini)
	{
	  //The global label of the rule set is the combination of the P system
          //name and its label in the particular P system
	  std::string secondLabel = ini->first;
	  std::string labelInitialMultiset = firstLabel + secondLabel;

	  combinedInitialMultisets[labelInitialMultiset] = ini->second;
	}
    }
  return combinedInitialMultisets;
}

unsigned long LPPsystem::getNumberOfCompartments()
{
  unsigned long numberCompartments = 0;

  for(std::map<std::string,PSystem>::iterator pSyst = (this->pSystems).begin();
      pSyst != (this->pSystems).end();
      ++pSyst)
    {
      numberCompartments += (pSyst->second).getNumberOfCompartments();
    }

  return numberCompartments;

}

//Displayer
void LPPsystem::show()
{

    std::cout << "----------------------------------------------\n";
    std::cout << "LPPsystem: " << this->name << std::endl << std::endl;

    std::cout << "List of P systems: " << std::endl;
    for(std::map<std::string,PSystem>::iterator pSys = this->pSystems.begin();
        pSys != this->pSystems.end();
        ++pSys)
    {
      (pSys->second).show();
    }

    this->lattice.show();

    this->positions.show();
}


/* Auxiliary functions for translating our SBML format into LPP format */
/* specially to perform model checking */

std::string LPPsystem::generateModel(const char* modelFile)
{
    generateLibrary(modelFile);
    std::string latticeFile = generateLattice(modelFile);
    std::map<std::string, std::vector< std::vector<std::string> > > positions = generatePsystems(modelFile);
    std::string lppFile = generateLPPsystem(modelFile,latticeFile,positions);
    return lppFile;
}

void LPPsystem::generateLibrary(const char* modelFile)
{
    //Name of the library file
    std::string fileName = string(modelFile);
    size_t pos = fileName.find(".");
    std::string libraryName = fileName.substr(0,pos);
    std::string fileLibraryName = libraryName + "_library.plb";

    ofstream libraryFile;
    libraryFile.open(fileLibraryName.c_str(),ios::out);

    //Determine the modules associated with the library
    std::map<std::string,std::vector<ECSB::LPPRule> > modules = obtainModules(modelFile);

    //Writing the library file
    libraryFile << "libraryOfModules library" << libraryName << std::endl << std::endl;

    for(std::map<std::string,std::vector<ECSB::LPPRule> >::iterator mod = modules.begin();
        mod != modules.end();
        ++mod)
    {
        libraryFile << "\t" << mod->first << "({},{},{}) =" << std::endl;
        libraryFile << "\t\t{\n";
        libraryFile << "\t\t\trules:\n";

        std::vector<ECSB::LPPRule> moduleRules = mod->second;

        for(std::vector<ECSB::LPPRule>::iterator r = moduleRules.begin();
            r != moduleRules.end();
            ++r)
        {
            libraryFile << "\t\t\t\t" << r->getName() << ":" << ruleString(*r) << std::endl;

        }


        libraryFile << "\t\t}\n\n";
    }

    libraryFile << "endLibraryOfModules\n";

    libraryFile.close();
}

std::map<std::string,std::vector<ECSB::LPPRule> > LPPsystem::obtainModules(const char* modelFile)
{
    std::map<std::string,std::vector<ECSB::LPPRule> > res;
    std::vector<std::string> modules;

    SBMLDocument* sbmlDocument = readSBML(modelFile);

    Model* sbmlModel = sbmlDocument->getModel();

    unsigned long numCompartments = sbmlModel->getNumCompartments();

    std::map<int,std::vector<ECSB::LPPRule> > rulesInModules;
    std::map<int,std::string> moduleNames = modulesName(modelFile);
    std::map<int,std::vector<int> > moduleReferences;

    //For each compartment we check if it represents a module. If so
    //the rules are stored in a map with key the integer id associated with
    //the module. In another map we store the other modules used to obtain
    //the rules of each module
    for(unsigned long comp = 0; comp < numCompartments; ++comp)
    {
        Compartment* sbmlCompartment = sbmlModel->getCompartment(comp);

        std::string compName = sbmlCompartment->getName();

        std::vector<std::string> nameParts = tokenizer(compName, ":");
        
        if(nameParts.size() < 4)
        {
            while(nameParts.size() < 4)
            {
                nameParts.push_back("");
            }
        }


        std::string moduleName;

        if(!compName.empty())
        {
            if(!nameParts[1].empty())
            {
                moduleName = nameParts[0] + "_" + nameParts[1];
                modules.push_back(moduleName);
                int intID = fromStringTo<int>(nameParts[1]);
            //moduleNames[intID] = moduleName;

                std::vector<ECSB::LPPRule> moduleRules =
                            obtainRules(sbmlCompartment->getId(),sbmlModel);

                rulesInModules[intID] = moduleRules;

                if(!nameParts[2].empty())
                {
                    std::vector<std::string> usedModules = tokenizer(nameParts[2],",");
                    std::vector<int> references;
                    for(std::vector<std::string>::iterator ref = usedModules.begin();
                        ref != usedModules.end();
                        ++ref)
                    {
                        references.push_back(fromStringTo<int>(*ref));
                    }

                    moduleReferences[fromStringTo<int>(nameParts[1])] = references;

                }

            }
        }

    }

    for(std::map<int,std::vector<int> >::iterator ref = moduleReferences.begin();
        ref != moduleReferences.end();
        ++ref)
    {
        for(std::vector<int>::iterator mod = (ref->second).begin();
            mod != (ref->second).end();
            ++mod)
        {
            std::vector<ECSB::LPPRule> rules = rulesInModules[*mod];

            for(std::vector<ECSB::LPPRule>::iterator r = rules.begin();
                r != rules.end();
                ++r)
            {
                rulesInModules[ref->first].push_back(*r);
            }
        }
    }

    for(std::map<int,std::string>::iterator mod = moduleNames.begin();
        mod != moduleNames.end();
        ++mod)
    {
        res[mod->second] = rulesInModules[mod->first];
    }

    return res;
}

std::map<int,std::string> LPPsystem::modulesName(const char* modelFile)
{
    SBMLDocument* sbmlDocument = readSBML(modelFile);
    Model* sbmlModel = sbmlDocument->getModel();

    unsigned long numCompartments = sbmlModel->getNumCompartments();
    std::map<int,std::string> names;

    for(unsigned long comp = 0; comp < numCompartments; ++comp)
    {
        Compartment* sbmlCompartment = sbmlModel->getCompartment(comp);
        std::string compName = sbmlCompartment->getName();
        std::vector<std::string> nameParts = tokenizer(compName, ":");

        if(nameParts.size() < 4)
        {
            while(nameParts.size() < 4)
            {
                nameParts.push_back("");
            }
        }


        std::string moduleName;

        if(!compName.empty() && !nameParts[1].empty())
        {
            moduleName = nameParts[0] + "_" + nameParts[1];
            int intID = fromStringTo<int>(nameParts[1]);
            names[intID] = moduleName;
        }
    }

    return names;

}

std::vector<ECSB::LPPRule> LPPsystem::obtainRules(std::string compId, Model *sbmlModel)
{

    std::vector<ECSB::LPPRule> rules;

    unsigned int numReactions = sbmlModel->getNumReactions();

    for(unsigned int r = 0; r < numReactions; ++r)
    {
        Reaction *reaction = sbmlModel->getReaction(r);
        std::string id = sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment();

        if(id == compId)
        {
            LPPRule rule;
            rule.setName(reaction->getId());
            rule.setLabel("cell");
            rule.setLabelLhs("cell");

            //The constant associated with the rule is the paramter associated with the kinetic law
            if(!reaction->isSetKineticLaw())
            {
                cerr << "No constant associated with reaction " << reaction->getId() << std::endl;
                exit(EXIT_FAILURE);
            }
            else if(reaction->getKineticLaw()->getNumParameters() == 0)
            {
                cerr << "No constant associated with reaction " << reaction->getId() << std::endl;
                exit(EXIT_FAILURE);
            }
            else if(reaction->getKineticLaw()->getNumParameters() > 1)
            {
                cerr << "More than a constant associated with reaction " << reaction->getId() << std::endl;
                exit(EXIT_FAILURE);
            }

            //rule.setStringConstant((reaction->getKineticLaw())->getParameter(0)->getName());
            rule.setStringConstant("c");
            rule.setConstant((reaction->getKineticLaw())->getParameter(0)->getValue());

            //Reactants and products
            std::vector<std::string> reactants = obtainReactants(sbmlModel,reaction);
            std::vector<std::string> products = obtainProducts(sbmlModel,reaction);

            std::string reactionName = reaction->getName();

            std::vector<std::string> parts = tokenizer(reactionName,":");

            if(parts.size() == 1)
            {
                rule.setType("boundary");
                rule.setLhsInside(reactants);
                rule.setRhsInside(products);
            }
            else if (parts.size() == 2)
            {
                rule.setType("translocation");
                rule.setLhsTranslocation(reactants);
                rule.setRhsTranslocation(products);

                std::vector<int> ruleVector;
                std::vector<std::string> vectorStr = tokenizer(parts[1],",");
                ruleVector.push_back(fromStringTo<int>(vectorStr[0]));
                ruleVector.push_back(fromStringTo<int>(vectorStr[1]));
                rule.setRuleVector(ruleVector);
            }

            rules.push_back(rule);
        }
    }

    return rules;
}

//Auxiliary function to determine the reactants of a rule from an SBML file
std::vector<std::string> LPPsystem::obtainReactants(Model* sbmlModel, Reaction* reaction)
{
    std::vector<std::string> reactantNames;
    unsigned int numReactants = reaction->getNumReactants();

    for(unsigned int s = 0; s < numReactants; ++s)
    {
        std::string speciesId = reaction->getReactant(s)->getSpecies();
        std::string speciesName = sbmlModel->getSpecies(speciesId)->getName();
        reactantNames.push_back(speciesName);
    }

    return reactantNames;
}

//Auxiliary function to test if species name is degradation species
bool LPPsystem::isDegradation(std::string s)
{
	if(s == "s0" || s.find("_degraded") != string::npos)
		return true;
	return false;
}

//Auxiliary function to determine the products of a rule from an SBML file
std::vector<std::string> LPPsystem::obtainProducts(Model* sbmlModel, Reaction* reaction)
{
    std::vector<std::string> productNames;
    unsigned int numProducts = reaction->getNumProducts();

    for(unsigned int s = 0; s < numProducts; ++s)
    {
        std::string speciesId = reaction->getProduct(s)->getSpecies();
        std::string speciesName = sbmlModel->getSpecies(speciesId)->getName();
        if(!isDegradation(speciesName))
        {
            productNames.push_back(speciesName);
        }
    }

    return productNames;
}

std::string LPPsystem::ruleString(ECSB::LPPRule rule)
{
    std::string ruleStr;

    if(rule.getType() == "boundary")
    {
        std::vector<std::string> reactants = rule.getLhsInside();
        std::vector<std::string> products = rule.getRhsInside();
        std::string label = rule.getLabel();
        std::string constantName = rule.getStringConstant();
        double constantValue = rule.getConstant();
        ruleStr = " [ ";

        if(reactants.size() == 1)
        {
            ruleStr += reactants[0];
        }
        else if(reactants.size() == 2)
        {
            ruleStr += reactants[0] + " + " + reactants[1];
        }
        else
        {
            cerr << "Not supported number of reactants in rule: " << std::endl;
            rule.show();
            exit(EXIT_FAILURE);
        }

        ruleStr += " ]_" + label + " -" + constantName + "-> [ ";

        if(products.size() == 1)
        {
            ruleStr += products[0];
        }
        else if(products.size() == 2)
        {
            ruleStr += products[0] + " + " + products[1];
        }
        else if (products.size() > 2)
        {
            cerr << "Not supported number of products in rule: " << std::endl;
            rule.show();
            exit(EXIT_FAILURE);
        }

        ruleStr += " ]_" + label + "\t" + constantName + " = " + somethingToString<double>(constantValue);

    }
    else if (rule.getType() == "translocation")
    {
        std::vector<std::string> reactants = rule.getLhsTranslocation();
        std::vector<std::string> products = rule.getRhsTranslocation();
        std::string label = rule.getLabel();
        std::string constantName = rule.getStringConstant();
        double constantValue = rule.getConstant();
        std::vector<int> ruleVector = rule.getVector();

        ruleStr = " [ ";

        if(reactants.size() == 1)
        {
            ruleStr += reactants[0];
        }
        else
        {
            cerr << "Not supproted number of reactants in rule: " << std::endl;
            rule.show();
            exit(EXIT_FAILURE);
        }

        ruleStr += " ]_" + rule.getLabel() + "=(";

        if(ruleVector.size() == 2)
        {
            ruleStr += somethingToString<int>(ruleVector[0]) + "," + somethingToString<int>(ruleVector[1]) + ")= [ ] ";
        }
        else
        {
            cerr << "Only 2D vectors are supported. Error in rule: " <<std::endl;
            rule.show();
            exit(EXIT_FAILURE);
        }

        ruleStr += "-" + constantName + "-> [ ]_" + rule.getLabel();
        ruleStr += " =(" + somethingToString<int>(ruleVector[0]) + "," + somethingToString<int>(ruleVector[1]) + ")= [ ";

        if(products.size() == 1)
        {
            ruleStr += products[0];
        }
        else
        {
            cerr << "Not supproted number of reactants in rule: " << std::endl;
            rule.show();
            exit(EXIT_FAILURE);
        }

        ruleStr += " ]    " + constantName + " = " + somethingToString<double>(constantValue); // + "\t" + constantName + " = " + somethingToString<double>(constantValue);

    }

    return ruleStr;
}

std::string LPPsystem::generateLattice(const char* modelFile)
{
    std::string fileName = string(modelFile);
    size_t pos = fileName.find(".");
    std::string latticeName = fileName.substr(0,pos);
    std::string fileLatticeName = latticeName + "_lattice.lat";

    ofstream latticeFile;
    latticeFile.open(fileLatticeName.c_str(),ios::out);

    std::vector<long> bounds = obtainBounds(modelFile);

    latticeFile << "lattice lattice_" << latticeName << std::endl << std::endl;

    latticeFile << "\tdimension\t" << 2 << std::endl;
    latticeFile << "\txmin\t\t\t" << bounds[0] << std::endl;
    latticeFile << "\txmax\t\t\t" << bounds[1] << std::endl;
    latticeFile << "\tymin\t\t\t" << bounds[2] << std::endl;
    latticeFile << "\tymax\t\t\t" << bounds[3] << std::endl << std::endl;

    latticeFile << "\tbasis" << std::endl;
    latticeFile << "\t\t(1,0)" << std::endl;
    latticeFile << "\t\t(0,1)" << std::endl;
    latticeFile << "\tendBasis" << std::endl << std::endl;

    latticeFile << "\tneighbours" << std::endl;
    latticeFile << "\t\t(1,0)" << std::endl;
    latticeFile << "\t\t(0,1)" << std::endl;
    latticeFile << "\t\t(-1,0)" << std::endl;
    latticeFile << "\t\t(0,-1)" << std::endl;
    latticeFile << "\tendNeighbours" << std::endl << std::endl;

    latticeFile << "endLattice" << std::endl;

    return fileLatticeName;

}

std::vector<long> LPPsystem::obtainBounds(const char* modelFile)
{
    //Bounds
    long xmin=0;
    long xmax=0;
    long ymin=0;
    long ymax=0;

    //Loading the sbml model
    SBMLDocument* sbmlDocument = readSBML(modelFile);

    Model* sbmlModel = sbmlDocument->getModel();

    //
    unsigned long numCompartments = sbmlModel->getNumCompartments();

    for(unsigned long comp = 0; comp < numCompartments; ++comp)
    {
        Compartment* sbmlCompartment = sbmlModel->getCompartment(comp);

        std::string compName = sbmlCompartment->getName();

        if(!compName.empty())
        {
            std::vector<std::string> nameParts = tokenizer(compName, ":");

            if(nameParts.size() < 4)
            {
                while(nameParts.size() < 4)
                {
                    nameParts.push_back("");
                }
            }


            /*

            std::cout << "vector for " << compName << std::endl;

            for(std::vector<std::string>::iterator i = nameParts.begin();
                i != nameParts.end();
                ++i)
            {
                std::cout << *i << std::endl;
            }
            */

            std::string position = nameParts[3];
            if(!position.empty())
            {
                std::vector<std::string> coordinates = tokenizer(position,",");
                long xcoord = fromStringTo<long>(coordinates[0]);
                long ycoord = fromStringTo<long>(coordinates[1]);

                if(xcoord < xmin)
                {
                    xmin = xcoord;
                }
                if(xcoord > xmax)
                {
                    xmax = xcoord;
                }
                if(ycoord < ymin)
                {
                    ymin = ycoord;
                }
                if(ycoord > ymax)
                {
                    ymax = ycoord;
                }
            }

        }
    }

    std::vector<long> bounds;

    bounds.push_back(xmin);
    bounds.push_back(xmax);
    bounds.push_back(ymin);
    bounds.push_back(ymax);

    return bounds;

}

std::map<std::string, std::vector< std::vector<std::string> > > LPPsystem::generatePsystems(const char* modelFile)
{
    std::map<std::string,std::string> pSystems;
    std::map<std::string, std::list<std::string> > pSystemAlphabets;
    std::map<std::string, std::vector<ECSB::LPPRule> > pSystemRules;
    std::map<std::string, std::map<std::string,int> > pSystemInitial;
    std::map<std::string, std::vector< std::vector<std::string> > > pSystemPositions;
    SBMLDocument* sbmlDocument = readSBML(modelFile);

    Model* sbmlModel = sbmlDocument->getModel();

    std::map<int,std::string> moduleNames = modulesName(modelFile);

    unsigned long numCompartments = sbmlModel->getNumCompartments();

    //Name of the library file
    std::string fileName = string(modelFile);
    size_t pos = fileName.find(".");
    std::string libraryName = fileName.substr(0,pos);
    std::string fileLibraryName = libraryName + "_library.plb";

    for(unsigned long comp = 0; comp < numCompartments; ++comp)
    {
        Compartment* sbmlCompartment = sbmlModel->getCompartment(comp);

        std::string compName = sbmlCompartment->getName();

        std::vector<std::string> nameParts = tokenizer(compName, ":");

        if(nameParts.size() < 4)
        {
            while(nameParts.size() < 4)
            {
                nameParts.push_back("");
            }
        }


        //If a position is associated with a compartment it represents a P system
        if(!compName.empty() && !nameParts[3].empty())
        {
            std::vector<std::string> positionStr = tokenizer(nameParts[3],",");
            pSystemPositions[nameParts[2]].push_back(positionStr);

            pSystems[nameParts[2]] = nameParts[0];
            pSystemAlphabets[nameParts[2]] = obtainAlphabet(sbmlCompartment,sbmlModel,nameParts[2]);
            pSystemRules[nameParts[2]] = obtainRules(sbmlCompartment->getId(),sbmlModel);
            pSystemInitial[nameParts[2]] = obtainInitial(sbmlModel,nameParts[2]);
        }
    }

    int pNumber = 0;
    std::map<std::string, std::vector< std::vector<std::string> > > positions;

    for(std::map<std::string,std::string>::iterator p = pSystems.begin();
        p != pSystems.end();
        ++p)
    {
        std::string pSystemName = p->second + "_" + somethingToString<int>(pNumber);
        ++pNumber;
        writePsystem(fileLibraryName,pSystemName,pSystemAlphabets[p->first],pSystemInitial[p->first],pSystemRules[p->first],p->first,moduleNames);
        positions[pSystemName] = pSystemPositions[p->first];
    }

/*

    for(std::map<std::string,std::string>::iterator pSyst = pSystems.begin();
        pSyst != pSystems.end();
        ++pSyst)
    {
        positions[pSyst->second] = pSystemPositions[pSyst->first];
    }
*/
    return positions;
}


std::list<std::string> LPPsystem::obtainAlphabet(Compartment* sbmlCompartment,
                                            Model* sbmlModel,
                                            std::string moduleStr)
{
    std::list<std::string> alphabet;

    std::vector<ECSB::LPPRule> rules = obtainRules(sbmlCompartment->getId(),sbmlModel);

    std::vector<std::string> modulesID = tokenizer(moduleStr,",");

    for(std::vector<std::string>::iterator m = modulesID.begin();
        m != modulesID.end();
        ++m)
    {
        std::string templateID = getTemplateID(*m,sbmlModel);
        std::vector<ECSB::LPPRule> templateRules = obtainRules(templateID, sbmlModel);
        rules.insert(rules.end(),templateRules.begin(),templateRules.end());
    }

    for(std::vector<ECSB::LPPRule>::iterator r = rules.begin();
        r != rules.end();
        ++r)
    {
        std::list<std::string> ruleObjects = r->getObjects();
        alphabet.splice(alphabet.end(),ruleObjects);
    }

    alphabet.sort();
    alphabet.unique();

    return alphabet;
}

std::map<std::string,int> LPPsystem::obtainInitial(Model* sbmlModel,
                                              std::string moduleStr)
{
    std::map<std::string,int> initials;

    std::vector<std::string> modulesID = tokenizer(moduleStr,",");

    for(std::vector<std::string>::iterator m = modulesID.begin();
        m != modulesID.end();
        ++m)
    {
        std::string templateID = getTemplateID(*m,sbmlModel);

        unsigned int numSpecies = sbmlModel->getNumSpecies();

        for(unsigned int i = 0; i < numSpecies; ++i)
        {
            Species* species = sbmlModel->getSpecies(i);
            std::string speciesComp = species->getCompartment();
            if(speciesComp == templateID)
            {
                double initialAmount = species->getInitialAmount();
                if(initialAmount != 0)
                {
                    initials[species->getName()] = initialAmount;
                }
            }
        }
    }

    return initials;
}

std::string LPPsystem::getTemplateID(std::string templateID, Model* sbmlModel)
{
    unsigned long numCompartments = sbmlModel->getNumCompartments();

    for(unsigned long comp = 0; comp < numCompartments; ++comp)
    {
        Compartment* sbmlCompartment = sbmlModel->getCompartment(comp);

        std::string compName = sbmlCompartment->getName();

        std::vector<std::string> nameParts = tokenizer(compName, ":");

        //If a position is associated with a compartment it represents a P system
        if(!compName.empty() && nameParts[1] == templateID)
        {
            return sbmlCompartment->getId();
        }
    }
}


void LPPsystem::writePsystem(std::string fileLibraryName,
                        std::string pSystemName,
                        std::list<std::string> alphabet,
                        std::map<std::string,int> initials,
                        std::vector<ECSB::LPPRule> rules,
                        std::string modulesID,
                        std::map<int,std::string> moduleNames)
{
    ofstream pSystemFile;
    std::string fileName = pSystemName + ".sps";
    pSystemFile.open(fileName.c_str(),ios::out);

    pSystemFile << "SPsystem " << pSystemName << std::endl << std::endl;


    pSystemFile << "\talphabet" << std::endl;

    for(std::list<std::string>::iterator obj = alphabet.begin();
        obj != alphabet.end();
        ++obj)
    {
        pSystemFile << "\t\t" << *obj << std::endl;
    }

    pSystemFile << "\tendAlphabet" << std::endl << std::endl;



    pSystemFile << "\tcompartments" << std::endl;
    pSystemFile << "\t\t" << pSystemName << std::endl;
    pSystemFile << "\tendCompartments" << std::endl << std::endl;

    pSystemFile << "\tinitialMultisets" << std::endl;

    pSystemFile << "\t\tinitialMultiset " << pSystemName << std::endl;

    for(std::map<std::string,int>::iterator sp = initials.begin();
        sp != initials.end();
        ++sp)
    {
        pSystemFile << "\t\t\t" << sp->first << " " << sp->second << std::endl;
    }

    pSystemFile << "\t\tendInitialMultiset" << std::endl;


    pSystemFile << "\tendInitialMultisets" << std::endl << std::endl;

    pSystemFile << "\truleSets" << std::endl << std::endl;

    pSystemFile << "\t\truleSet " << pSystemName << std::endl;

    int ruleID = 1;

    for(std::vector<ECSB::LPPRule>::iterator r = rules.begin();
        r != rules.end();
        ++r)
    {
        pSystemFile << "\t\t\tr_" << ruleID << ": " << ruleString(*r) << std::endl;
        ruleID ++;
    }

    std::vector<std::string> modules = tokenizer(modulesID,",");

    for(std::vector<std::string>::iterator m = modules.begin();
        m != modules.end();
        ++m)
    {
        pSystemFile << "\t\t\t" << moduleNames[fromStringTo<int>(*m)];
        pSystemFile << "({},{},{})" << " from " << fileLibraryName << std::endl;
    }

    pSystemFile << "\t\tendRuleSet" << std::endl << std::endl;

    pSystemFile << "\tendRuleSets" << std::endl << std::endl;

    pSystemFile << "endSPsystem" << std::endl;

    pSystemFile.close();

}


std::string LPPsystem::generateLPPsystem(const char* modelFile,
                             std::string latticeFile,
                             std::map<std::string, std::vector< std::vector<std::string> > > positions)
{
    //Name of the LPP-system
    std::string fileName = string(modelFile);
    size_t pos = fileName.find(".");
    std::string lppSystemName = fileName.substr(0,pos);
    std::string fileLPPSystemName = lppSystemName + ".lpp";

    ofstream fileLPPSystem;
    fileLPPSystem.open(fileLPPSystemName.c_str(),ios::out);

    fileLPPSystem << "LPPsystem " << lppSystemName << std::endl << std::endl;
    fileLPPSystem << "\tSPsystems" << std::endl;

    for(std::map<std::string, std::vector< std::vector<std::string> > >::iterator pSyst = positions.begin();
        pSyst != positions.end();
        ++pSyst)
    {
        fileLPPSystem << "\t\tSPsystem " << pSyst->first << " from " << pSyst->first << ".sps" << std::endl;
    }

    fileLPPSystem << "\tendSPsystems" << std::endl << std::endl;


    size_t posDot = latticeFile.find(".");
    std::string latticeName = latticeFile.substr(0,posDot);

    fileLPPSystem << "\tlattice " << latticeName << " from " << latticeFile << std::endl << std::endl;

    fileLPPSystem << "\tspatialDistribution" << std::endl << std::endl;

    for(std::map<std::string, std::vector< std::vector<std::string> > >::iterator pSyst = positions.begin();
        pSyst != positions.end();
        ++pSyst)
    {
        std::vector< std::vector<std::string> > pSystemPositions = pSyst->second;

        for(std::vector< std::vector<std::string> >::iterator pos = pSystemPositions.begin();
            pos != pSystemPositions.end();
            ++pos)
        {
            std::vector<std::string> position = *pos;

            fileLPPSystem << "\t\tpositions for " << pSyst->first << std::endl;
            fileLPPSystem << "\t\t\tcoordinates" << std::endl;
            fileLPPSystem << "\t\t\t\tx = " << position[0] << std::endl;
            fileLPPSystem << "\t\t\t\ty = " << position[1] << std::endl;
            fileLPPSystem << "\t\t\tendCoordinates" << std::endl;
            fileLPPSystem << "\t\tendPositions" << std::endl << std::endl;
        }
    }

    fileLPPSystem << "\tendSpatialDistribution" << std::endl << std::endl;

    fileLPPSystem << "endLPPsystem";

    fileLPPSystem.close();

    return fileLPPSystemName;

}











