/*
 * File:   PSystem.cpp
 * Author: Francisco J. Romero-Campero
 *
 * Created on 15 January 2010, 16:31
 */


#include "PSystem.h"

using namespace std;
using namespace ECSB;
using namespace xmlpp;


//Default constructor
PSystem::PSystem()
{

}

//Copy constructor
PSystem::PSystem(const PSystem& orig)
{
    name = orig.name;
    alphabet = orig.alphabet;
    setOfLabels = orig.setOfLabels;
    membraneStructure = orig.membraneStructure;
    initialMultisets = orig.initialMultisets;
    ruleSets = orig.ruleSets;
}

//Default destructor
PSystem::~PSystem()
{

}

//Constructor from a file
PSystem::PSystem(std::string fileName)
{
    //Getting the extesion of the file
    size_t pos = fileName.find('.');
    std::string fileExtension = fileName.substr(pos+1);

    if(fileExtension == "sps")
    {
        //Varibale for the xml file with the definition of a lattice
        std::string xmlFile;

        //The P system parser generated from our BNF grammar
        Psystem_p pParser;    //  Our parser

        //Generate a string with the content of the file
        std::string stringFile = fileToString(fileName);

        //We parse the content of the file to generate an abstract syntax tree
        tree_parse_info<> result = ast_parse(stringFile.c_str(),
                                             pParser,
                                             space_p);

        if (result.full) //When parsing is succesfull the xml file is generated
        {
            iter_t iter = result.trees.begin();
            xmlFile = generatePSystemXML(iter);
        }
        else //Failed parsing
        {
            std::map<unsigned long, std::string> fileLines = fileToLines(fileName);
            unsigned long lineNumber = findLine(result.stop,fileLines);

            cerr << "Parsing of file " << fileName;
            cerr << " according to BNF failed in line: " << lineNumber << std::endl;
            //std::cout << result.stop << std::endl;
            exit(EXIT_FAILURE);
        }

        //Create the P system using the xml file
        createFromXML(xmlFile);
    }
    else if(fileExtension == "xml")
    {
        createFromXML(fileName);
    }
    else if(fileExtension == "sbml")
    {
        createFromSBML(fileName);
    }
    else
    {
        cerr << "Error: unknown file extension in file: " << fileName << std::endl;
        cerr << "Formats supported:\n" ;
        cerr << "\t SBML: file must have .sbml extension \n";
        cerr << "\t SPS: text file must be specified according to our BNF \n";
        cerr << "\t      and the fiel extension must be .sps";
        exit(EXIT_FAILURE);
    }

    //show();
    //std::string pause;
    //cin >> pause;

}

void PSystem::createFromXML(std::string fileName)
{
    //Use of the DOM parser
    DomParser parser;
    parser.parse_file(fileName);

    if(!parser)
    {
        cerr << "Parsing unsuccesfull for file " << fileName << std::endl;
        exit(EXIT_FAILURE);
    }

    xmlpp::Node* pSystemDeclaration = parser.get_document()->get_root_node();

    //Name of the PSystem
    xmlpp::Element* pSystemElement = dynamic_cast<xmlpp::Element*>(pSystemDeclaration);
    this->name = (pSystemElement->get_attribute("name"))->get_value();

    //Get the declaration of the different components
    xmlpp::Node::NodeList components = pSystemDeclaration->get_children();
    xmlpp::Node::NodeList alphabetDeclaration;
    xmlpp::Node::NodeList labelsDeclaration;
    xmlpp::Node::NodeList compartmentsDeclaration;
    xmlpp::Node::NodeList initialMultisetsDeclaration;
    xmlpp::Node::NodeList ruleSetsDeclaration;

    for(xmlpp::Node::NodeList::iterator comp = components.begin(); comp != components.end(); ++comp)
      {
        if((*comp)->get_name() == "alphabetOfObjects")
	{
	  alphabetDeclaration = (*comp)->get_children("object");
	}
      else if ((*comp)->get_name() == "compartmentLabels")
	{
	  labelsDeclaration = (*comp)->get_children("compartmentLabel");
	}
      else if ((*comp)->get_name() == "membraneStructure")
	{
	  compartmentsDeclaration = (*comp)->get_children("compartment");
	}

      else if ((*comp)->get_name() == "initialMultisets")
	{
	  initialMultisetsDeclaration = (*comp)->get_children("initialMultiset");
	}
      else if ((*comp)->get_name() == "ruleSets")
	{
	  ruleSetsDeclaration = (*comp)->get_children("ruleSet");
	}
    }

  //Construct the different components from their declarations
  createAlphabet(alphabetDeclaration);
  createSetOfLabels(labelsDeclaration);
  createMembraneStructure(compartmentsDeclaration);
  assignChildren();
  createInitialMultisets(initialMultisetsDeclaration);
  createRuleSets(ruleSetsDeclaration);
  checkAlphabet();
}

//Auxiliary function to create the alphabet from an XML file
void PSystem::createAlphabet(xmlpp::Node::NodeList alphabetDeclaration)
{
  for(xmlpp::Node::NodeList::iterator obj = alphabetDeclaration.begin();
      obj != alphabetDeclaration.end();
      ++obj)
    {
      //We get the name of each object and store it
      xmlpp::Element* objElement = dynamic_cast<xmlpp::Element*>(*obj);
      std::string objName = (objElement->get_attribute("name"))->get_value();
      (this->alphabet).push_back(objName);
    }
}

//Auxiliary function to create the lables from an XML file
void PSystem::createSetOfLabels(xmlpp::Node::NodeList labelsDeclaration)
{
  for(xmlpp::Node::NodeList::iterator lab = labelsDeclaration.begin();
      lab != labelsDeclaration.end();
      ++lab)
    {
      //We get the name of each label and store it
      xmlpp::Element* labElement = dynamic_cast<xmlpp::Element*>(*lab);
      std::string labName = (labElement->get_attribute("label"))->get_value();
      (this->setOfLabels).push_back(labName);
    }
}

//Auxiliary function to create the membrane structure from an XML file
void PSystem::createMembraneStructure(xmlpp::Node::NodeList compartmentsDeclaration)
{
    LPPCompartment compartment;
    unsigned long numberId = 0;

    for(xmlpp::Node::NodeList::iterator comp = compartmentsDeclaration.begin();
       comp != compartmentsDeclaration.end();
       ++comp)
    {
      compartment = LPPCompartment(*comp);
      compartment.setNumberId(numberId);
      numberId++;

      //(this->membraneStructure).insert(pair<std::string,CAPCompartment>(compartment.getName(),compartment));
      (this->membraneStructure)[compartment.getName()] = compartment;
    }
}

//Auxiliary function to determine the compartments inside a given one
std::vector<std::string> PSystem::findChildren(std::string compartmentName)
{
    std::vector<std::string> childrenNames;

    for(std::map<std::string,LPPCompartment>::iterator comp = this->membraneStructure.begin();
    comp != this->membraneStructure.end();
    ++comp)
    {
        if((comp->second).getFather() == compartmentName)
        {
            childrenNames.push_back((comp->second).getName());
        }
    }

    return childrenNames;
}

//Auxiliary function to assign the children of each compartment
void PSystem::assignChildren()
{
    for(std::map<std::string,LPPCompartment>::iterator comp = this->membraneStructure.begin();
    comp != this->membraneStructure.end();
    ++comp)
    {
        (comp->second).setChildren(findChildren((comp->second).getName()));
    }
}

//Auxiliary function to create the initial multisets from an XML file
void PSystem::createInitialMultisets(xmlpp::Node::NodeList initialMultisetsDeclaration)
{
  for(xmlpp::Node::NodeList::iterator ini = initialMultisetsDeclaration.begin();
      ini != initialMultisetsDeclaration.end();
      ++ini)
    {
      InitialMultiset initial(*ini);
      (this->initialMultisets)[initial.getLabel()] = initial;
      //(this->initialMultisets).insert(pair<std::string,InitialMultiset>(i.getLabel(),i));
    }
}

//Auxiliary function to create the rule sets from an XML file
void PSystem::createRuleSets(xmlpp::Node::NodeList ruleSetsDeclaration)
{
  for(xmlpp::Node::NodeList::iterator rul = ruleSetsDeclaration.begin();
      rul != ruleSetsDeclaration.end();
      ++rul)
    {
      LPPRuleSet rs(*rul);
      (this->ruleSets)[rs.getLabel()] = rs;
      //(this->ruleSets).insert(pair<std::string,LPPRuleSet>(rs.getLabel(),rs));
  }
}

//Auxiliary function to check if all the objects used in the rules are
//specified in the alphabet
void PSystem::checkAlphabet()
{
    std::list<std::string> ruleObjects = getRuleObjects();
    std::list<std::string> alphabetObjects((this->alphabet).begin(),(this->alphabet).end());
    alphabetObjects.sort();
    alphabetObjects.unique();

    if(ruleObjects != alphabetObjects)
    {
        std::cout << "Warning missing objects in the alphabet from P system " << this->name << std::endl;
        std::cout << "Suggested Alphabet:" << std::endl;
        std::vector<std::string> newAlphabet;
        std::cout << "\talphabet" << std::endl;
        for(std::list<std::string>::iterator o = ruleObjects.begin();
            o != ruleObjects.end();
            ++o)
        {
            std::cout << "\t\t" << *o << std::endl;
            newAlphabet.push_back(*o);
        }
        std::cout << "\tendAlphabet" << std::endl << std::endl;

        std::cout << "Do you want to change the alphabet? (y/n): ";
        std::string answer;
        std::cin >> answer;
        if(answer == "y")
        {
            this->alphabet = newAlphabet;
        }
        else
        {
            exit(EXIT_FAILURE);
        }
    }
}

//Auxiliary function to collect of the objects used in the rules
std::list<std::string> PSystem::getRuleObjects()
{
    std::list<std::string> objects;

    for(std::map<std::string,LPPRuleSet>::iterator rSet = (this->ruleSets).begin();
    rSet != (this->ruleSets).end();
    ++rSet)
    {
        std::list<std::string> ruleSetObjects = (rSet->second).getObjects();

        std::list<std::string>::iterator it = objects.begin();
        objects.splice(it,ruleSetObjects);
    }

    objects.sort();
    objects.unique();

    return objects;
}


//Auxiliary function that converts all the rules in the P system that move
//objects from the ouside to the inside of comparments and viceversa in the
//hierarchical structure of membranes into a translocation rule
void PSystem::convertTranslocation()
{
    for(std::map<std::string,LPPRuleSet>::iterator rSet = this->ruleSets.begin();
        rSet != this->ruleSets.end();
        ++rSet)
    {
        std::vector<LPPRule> convertedRules;
        std::vector<LPPRule> rules = (rSet->second).getRules();

        for(std::vector<LPPRule>::iterator rule = rules.begin();
        rule != rules.end();
        ++rule)
        {
            convertedRules.push_back(convertTranslocationRule(*rule, rSet->first));
        }

        (rSet->second).setRules(convertedRules);
    }
}

//Auxiliary function that converts a rule that moves objects from the ouside
//to the inside of comparments and viceversa in the hierarchical structure of
//membranes into a translocation rule
LPPRule PSystem::convertTranslocationRule(LPPRule rule,
                                          std::string compartmentName)
{
    LPPRule convertedRule;

    //Rule of the form [ o1 ] -> [ o2] or translocation
    if((rule.getLhsOutside().empty() && rule.getRhsOutside().empty()) |
       (rule.getType() == "translocation"))
    {
        convertedRule = rule;
    }
    //[ A ]_name -> A [ ]_name
    else if(rule.getLhsOutside().empty() &&
            !rule.getLhsInside().empty() &&
            !rule.getRhsOutside().empty() &&
            rule.getRhsInside().empty())
    {
        convertedRule.setConstant(rule.getConstant());
        convertedRule.setLabel(rule.getLabel());
        convertedRule.setLabelLhs(rule.getLabel());
        convertedRule.setName(rule.getName());
        convertedRule.setStringConstant(rule.getStringConstant());
        convertedRule.setType("translocation");

        convertedRule.setLhsTranslocation(rule.getLhsInside());
        convertedRule.setRhsTranslocation(rule.getRhsOutside());

        //Determining the rule vector
        int sourceId = (this->membraneStructure)[compartmentName].getNumberId();
        std::string fatherCompartment = (this->membraneStructure)[compartmentName].getFather();
        int targetId = (this->membraneStructure)[fatherCompartment].getNumberId();
        //The first two components are zero, only the third one is used to
        //represent compartments inside compartments
        std::vector<int> ruleVector(2,0);
        //Computing the third component
        ruleVector.push_back(targetId - sourceId);
        convertedRule.setRuleVector(ruleVector);
    }
    //A [ ]_name -> [ A ]_name
    else if(!rule.getLhsOutside().empty() &&
            rule.getLhsInside().empty() &&
            rule.getRhsOutside().empty() &&
            !rule.getRhsInside().empty())
    {
        convertedRule.setConstant(rule.getConstant());
        convertedRule.setLabel(rule.getLabel());
        convertedRule.setLabelLhs(rule.getLabel());
        convertedRule.setName(rule.getName());
        convertedRule.setStringConstant(rule.getStringConstant());
        convertedRule.setType("translocation");

        convertedRule.setLhsTranslocation(rule.getLhsOutside());
        convertedRule.setRhsTranslocation(rule.getRhsInside());

        //Determining the rule vector
        unsigned long sourceId = (this->membraneStructure)[compartmentName].getNumberId();
        unsigned long targetId = (this->membraneStructure)[rule.getLabel()].getNumberId();
        std::vector<int> ruleVector(2,0);
        ruleVector.push_back(targetId - sourceId);
        convertedRule.setRuleVector(ruleVector);
    }
    else
    {
        cerr << "Rule type no suported: " << std::endl;
        rule.show();
        exit(EXIT_FAILURE);
    }

    return convertedRule;
}

//Function that adds a rule to a rule set if the latter one doesn't exist it
//creates a new one and adds the rule
void PSystem::addRule(std::string compartmentName, LPPRule rule)
{
    std::map<std::string,LPPRuleSet>::iterator it = this->ruleSets.find(compartmentName);

    if(it == this->ruleSets.end())
    {
        LPPRuleSet ruleSet;
        ruleSet.setLabel(compartmentName);
        ruleSet.addRule(rule);
        this->ruleSets[ruleSet.getLabel()] = ruleSet;
        //this->ruleSets.insert(pair<std::string,LPPRuleSet>(ruleSet.getLabel(),ruleSet));
    }
    else
    {
        (it->second).addRule(rule);
    }
}

//Auxiliary function to create a P system from an SBML file
void PSystem::createFromSBML(std::string fileName)
{
    //Read the sbml file
    SBMLDocument* sbmlDocument = readSBML(fileName.c_str());

    //Checking for errors
    unsigned int numErrors = sbmlDocument->getNumErrors();
    if(numErrors > 0)
    {
        cerr << "Errors when reading sbml file" << std::endl;
	sbmlDocument->printErrors(cerr);
	exit(EXIT_FAILURE);
    }

    if(sbmlDocument->getLevel() == 1)
    {
        cerr << "Error when reading sbml file: incompatible version" << std::endl;
	exit(EXIT_FAILURE);
    }

    Model* sbmlModel = sbmlDocument->getModel();

    //Set model name
    this->name = sbmlModel->getId();

    //Construct the different components
    createAlphabet(sbmlModel);
    createMembraneStructure(sbmlModel);
    assignChildren();
    createSetOfLabels();
    createInitialMultisets(sbmlModel);
    createRuleSets(sbmlModel);
}

//Auxiliary function to create the alphabet from an SBML file
void PSystem::createAlphabet(Model* sbmlModel)
{
    std::list<std::string> speciesNames;
    unsigned int numSpecies = sbmlModel->getNumSpecies();

    for(unsigned int s = 0; s < numSpecies; ++s)
    {
        std::string speciesName = sbmlModel->getSpecies(s)->getName();
        if(!isDegradation(speciesName))
        {
            speciesNames.push_front(speciesName);
        }
    }

    speciesNames.sort();
    speciesNames.unique();

    std::vector<std::string> species(speciesNames.begin(),speciesNames.end());

    this->alphabet = species;
}

//Auxiliary function to test if species name is degradation species
bool PSystem::isDegradation(std::string s)
{
	if(s == "s0" || s.find("_degraded") != string::npos)
		return true;
	return false;
}

//Auxiliary function to create the membrane structure from an SBML file
void PSystem::createMembraneStructure(Model* sbmlModel)
{
    unsigned int numCompartments = sbmlModel->getNumCompartments();
    unsigned int numberId = 0;

    for(unsigned int c = 0; c < numCompartments; ++c)
    {
        LPPCompartment compartment;
        Compartment* sbmlCompartment = sbmlModel->getCompartment(c);

        //The default compartment is ignored
        if(sbmlCompartment->getId() != "default")
        {
            compartment.setName(sbmlCompartment->getName());
            compartment.setLabel(sbmlCompartment->getName());
            compartment.setNumberId(numberId);
            ++numberId;
            std::string fatherId = sbmlCompartment->getOutside();
            Compartment* fatherCompartment = sbmlModel->getCompartment(fatherId);
            compartment.setFather(fatherCompartment->getName());
            (this->membraneStructure)[compartment.getName()]=compartment;
            //(this->membraneStructure).insert(pair<std::string,CAPCompartment>(compartment.getName(),compartment));
        }
    }
}

//Auxiliary function to create the set of labels from an SBML file
void PSystem::createSetOfLabels()
{
    for(std::map<std::string,LPPCompartment>::iterator comp = this->membraneStructure.begin();
    comp != this->membraneStructure.end();
    ++comp)
    {
        this->setOfLabels.push_back(comp->first);
    }
}

//Auxiliary function to create the initial multisets from an SBML model
void PSystem::createInitialMultisets(Model* sbmlModel)
{
    unsigned int numSpecies = sbmlModel->getNumSpecies();

    for(unsigned int s = 0; s < numSpecies; ++s)
    {
        std::string speciesName = sbmlModel->getSpecies(s)->getName();
        unsigned int multiplicity = sbmlModel->getSpecies(s)->getInitialAmount();
        if(multiplicity > 0)
        {
            std::string compartmentId = sbmlModel->getSpecies(s)->getCompartment();
            std::string compartmentName = sbmlModel->getCompartment(compartmentId)->getName();
            std::map<std::string,InitialMultiset>::iterator it = this->initialMultisets.find(compartmentName);
            //If the initial multisets has not been created yet it is created and
            //initialised witht the current multiplicity
            if(it == this->initialMultisets.end())
            {
                InitialMultiset iniMultiset;
                iniMultiset.setLabel(compartmentName);
                iniMultiset.setMultiplicity(speciesName,multiplicity);
                this->initialMultisets.insert(pair<std::string,InitialMultiset>(compartmentName,iniMultiset));
            }
            else
            {
                (it->second).setMultiplicity(speciesName,multiplicity);
            }
        }
    }
}

//Auxiliary function to cerate the sets of rules from an SBML file
void PSystem::createRuleSets(Model* sbmlModel)
{
    unsigned int numReactions = sbmlModel->getNumReactions();

    for(unsigned int r = 0; r < numReactions; ++r)
    {
        unsigned int numReactants = sbmlModel->getReaction(r)->getNumReactants();
        Reaction* reaction = sbmlModel->getReaction(r);

        //Checking if the reaction format is supported
        checkValid(sbmlModel,reaction);

        //Create a rule from the corresponding reaction
        LPPRule rule;

        rule.setName(reaction->getId());

        //The rule is associated with the compartment where the reactants are located
        std::string compartmentId = sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment();
        std::string compartmentName = sbmlModel->getCompartment(compartmentId)->getName();

        //The constant associated with the rule is the paramter associated with the kinetic law
        if(!reaction->isSetKineticLaw())
        {
            cerr << "No constant associated with reaction " << reaction->getId() << std::endl;
            exit(EXIT_FAILURE);
        }
        else if(reaction->getKineticLaw()->getNumParameters() == 0)
        {
            cerr << "No constant associated with reaction " << reaction->getId() << std::endl;
            exit(EXIT_FAILURE);
        }
        else if(reaction->getKineticLaw()->getNumParameters() > 1)
        {
            cerr << "More than a constant associated with reaction " << reaction->getId() << std::endl;
            exit(EXIT_FAILURE);
        }

        rule.setStringConstant((reaction->getKineticLaw())->getParameter(0)->getName());
        rule.setConstant((reaction->getKineticLaw())->getParameter(0)->getValue());

        //Getting the reactants and products of the reaction to specify the lhs and rhs of the rule
        std::vector<std::string> lhsObjects = reactants(sbmlModel, reaction);
        std::vector<std::string> rhsObjects = products(sbmlModel, reaction);

        //Depending on the type of reaction the rule is created accordingly
        std::string type = reactionType(sbmlModel, reaction);
        if(type == "transformation")
        {
            rule.setType("boundary");
            rule.setLhsInside(lhsObjects);
            rule.setRhsInside(rhsObjects);
            rule.setLabel(sbmlModel->getCompartment(sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment())->getName());
            addRule(compartmentName,rule);
        }
        else if(type == "diffusionOut")
        {
            rule.setType("boundary");
            rule.setLhsInside(lhsObjects);
            rule.setRhsOutside(rhsObjects);
            rule.setLabel(sbmlModel->getCompartment(sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment())->getName());
            addRule(compartmentName,rule);
        }
        else if(type == "diffusionIn")
        {
            rule.setType("boundary");
            rule.setLhsOutside(lhsObjects);
            rule.setRhsInside(rhsObjects);
            rule.setLabel(sbmlModel->getCompartment(sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment())->getName());
            addRule(compartmentName,rule);
        }
        else if(type == "translocation")
        {
            //The diffusion out of the system is represented as a translocation rule sending objects to the neighbours of the P system
            //in the lattice. We assume a rectangular lattice with only four neighbours this needs to be made general.
            rule.setType("translocation");
            rule.setLhsTranslocation(lhsObjects);
            rule.setRhsTranslocation(rhsObjects);
            rule.setLabel(sbmlModel->getCompartment(sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment())->getName());

            rule.setLabelLhs(sbmlModel->getCompartment(sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment())->getName());

            //We add four rules so the constant rule of each new rule need to be reajusted in order not to disturb the dynamics of the system.
            double ruleConstant = rule.getConstant();
            ruleConstant = ruleConstant / 4;
            rule.setConstant(ruleConstant);

            std::vector<int> ruleVector(3,0);

            LPPRule ruleW = rule;
            ruleVector[0] = -1;
            ruleVector[1] = 0;
            ruleW.setRuleVector(ruleVector);
            addRule(compartmentName,ruleW);

            LPPRule ruleE = rule;
            ruleVector[0] = 1;
            ruleVector[1] = 0;
            ruleE.setRuleVector(ruleVector);
            addRule(compartmentName,ruleE);

            LPPRule ruleN = rule;
            ruleVector[0] = 0;
            ruleVector[1] = 1;
            ruleN.setRuleVector(ruleVector);
            addRule(compartmentName,ruleN);

            LPPRule ruleS = rule;
            ruleVector[0] = 0;
            ruleVector[1] = -1;
            ruleS.setRuleVector(ruleVector);
            addRule(compartmentName,ruleS);

        }
    }
}

//Auxiliary function to check if the format of a rule is supported
void PSystem::checkValid(Model* sbmlModel, Reaction* reaction)
{
    unsigned int numReactants = reaction->getNumReactants();
    unsigned int numProducts = reaction->getNumProducts();

    //We only support reactions with at least one reactant and one product and no more than two reactants and two products.
    if(numReactants == 0 || numReactants > 2)
    {
        cerr << "Invalid number of reactants:" << numReactants << " in rule " << reaction->getId() << std::endl;
        exit(EXIT_FAILURE);
    }

    if(numProducts == 0 || numProducts > 2)
    {
        cerr << "Invalid number of products: " << numProducts << " in rule " << reaction->getId() << std::endl;
        exit(EXIT_FAILURE);
    }

    //All the reactants and products must be in the same compartment if we have either two reactants or two products
    if(numReactants == 2 && numProducts == 1)
    {
        std::string compartmentName1 = sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment();
        std::string compartmentName2 = sbmlModel->getSpecies(reaction->getReactant(1)->getSpecies())->getCompartment();
        std::string compartmentName3 = sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment();
        if(compartmentName1 != compartmentName2 || compartmentName1 != compartmentName3)
        {
            cerr << "Reactants and products in different compartments in rule " << reaction->getId() << std::endl;
            exit(EXIT_FAILURE);
        }
    }

    if(numReactants == 2 && numProducts == 2)
    {
        std::string compartmentName1 = sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment();
        std::string compartmentName2 = sbmlModel->getSpecies(reaction->getReactant(1)->getSpecies())->getCompartment();
        std::string compartmentName3 = sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment();
        std::string compartmentName4 = sbmlModel->getSpecies(reaction->getProduct(1)->getSpecies())->getCompartment();
        if(compartmentName1 != compartmentName2 || compartmentName1 != compartmentName3 || compartmentName1 != compartmentName4)
        {
            cerr << "Reactants and products in different compartments in rule " << reaction->getId() << std::endl;
            exit(EXIT_FAILURE);
        }
    }

    if(numReactants == 1 && numProducts == 2)
    {
        std::string compartmentName1 = sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment();
        std::string compartmentName2 = sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment();
        std::string compartmentName3 = sbmlModel->getSpecies(reaction->getProduct(1)->getSpecies())->getCompartment();
        if(compartmentName1 != compartmentName2 || compartmentName1 != compartmentName3)
        {
            cerr << "Reactants and products in different compartments in rule " << reaction->getId() << std::endl;
            exit(EXIT_FAILURE);
        }
    }
}

//Auxiliary function to determine the type of the rule
std::string PSystem::reactionType(Model* sbmlModel, Reaction* reaction)
{
    std::string type;

    unsigned int numReactants = reaction->getNumReactants();
    unsigned int numProducts = reaction->getNumProducts();

    if(numReactants == 2 || numProducts == 2)
    {
        type = "transformation";
    }

    if(numReactants == 1 && numProducts == 1)
    {
        std::string reactantCompartment = sbmlModel->getSpecies(reaction->getReactant(0)->getSpecies())->getCompartment();
        std::string reactantOutside = sbmlModel->getCompartment(reactantCompartment)->getOutside();
        std::string productCompartment = sbmlModel->getSpecies(reaction->getProduct(0)->getSpecies())->getCompartment();
        std::string productOutside = sbmlModel->getCompartment(productCompartment)->getOutside();

        if(reactantOutside == productCompartment)
        {
            if(reactantOutside == "default")
            {
                type = "translocation";
            }
            else
            {
                type = "diffusionOut";
            }
        }
        else if(reactantCompartment == productOutside)
        {
            type = "diffusionIn";
        }
        else if(reactantCompartment == productCompartment)
        {
            type = "transformation";
        }
    }

    return type;
}

//Auxiliary function to determine the reactants of a rule from an SBML file
std::vector<std::string> PSystem::reactants(Model* sbmlModel, Reaction* reaction)
{
    std::vector<std::string> reactantNames;
    unsigned int numReactants = reaction->getNumReactants();

    for(unsigned int s = 0; s < numReactants; ++s)
    {
        std::string speciesId = reaction->getReactant(s)->getSpecies();
        std::string speciesName = sbmlModel->getSpecies(speciesId)->getName();
        reactantNames.push_back(speciesName);
    }

    return reactantNames;
}

//Auxiliary function to determine the products of a rule from an SBML file
std::vector<std::string> PSystem::products(Model* sbmlModel, Reaction* reaction)
{
    std::vector<std::string> productNames;
    unsigned int numProducts = reaction->getNumProducts();

    for(unsigned int s = 0; s < numProducts; ++s)
    {
        std::string speciesId = reaction->getProduct(s)->getSpecies();
        std::string speciesName = sbmlModel->getSpecies(speciesId)->getName();
        if(!isDegradation(speciesName))
        {
            productNames.push_back(speciesName);
        }
    }

    return productNames;
}

//Displayer
void PSystem::show()
{
  std::cout << "-----------------------------" << std::endl;
  std::cout << "P System: " << this->name << std::endl;
  std::cout << "-----------------------------" << std::endl;

  //Show the alphabet
  std::cout << "Alphabet: " << std::endl;
  for(std::vector<std::string>::iterator obj = this->alphabet.begin();
      obj != this->alphabet.end();
      ++obj)
    {
      std::cout << "\t" << *obj << std::endl;
    }

  //Show the labels
  std::cout << "Labels: " << std::endl;
  for(std::vector<std::string>::iterator lab = this->setOfLabels.begin();
      lab != this->setOfLabels.end();
      ++lab)
    {
      std::cout << "\t" << *lab << std::endl;
    }

  //Show membrane structure
  std::cout << "Membrane Structure:" << std::endl;
  for(std::map<std::string,LPPCompartment>::iterator comp = this->membraneStructure.begin();
      comp != this->membraneStructure.end();
      ++comp)
    {
      (comp->second).show();
    }
  //Show initial multisets
  std::cout << "Initial Multisets: " << std::endl;
  for(std::map<std::string,InitialMultiset>::iterator ini = this->initialMultisets.begin();
      ini != this->initialMultisets.end();
      ++ini)
    {
      (ini->second).show();
    }
  //Show set of rules
  std::cout << "Set of Rules " << std::endl;
  for(std::map<std::string,LPPRuleSet>::iterator rSet = ruleSets.begin();
      rSet != ruleSets.end();
      ++rSet)
    {
      (rSet->second).show();
    }
}
