
/*
 * File:   PSystemBNF.cpp
 * Author: Francisco J. Romero-Campero
 *
 * Created on 14 January 2010, 12:27
 */

#include "PSystemBNF.h"


using namespace std;
using namespace ECSB;
using namespace xmlpp;
using namespace boost::spirit::classic;

//Function to generate an XML file for the specification of a P system from
//a text file in accordance with our BNF grammar
std::string ECSB::generatePSystemXML(iter_t tree)
{
    //Create xml file and root node
    xmlpp::Document xmlFile;
    xmlpp::Element* psystemNode = xmlFile.create_root_node("PSystem");

    //Associated p system name
    std::string psystemName(tree->value.begin(), tree->value.end());
    psystemName = removeSpaces(psystemName);
    psystemNode->set_attribute("name",psystemName);

    //Add children corresponding to the alphabet, membrane structure, initial
    //multisets and rule sets
    xmlpp::Element* alphabetNode =
            psystemNode->add_child("alphabetOfObjects");
    xmlpp::Element* compartmentLabelsNode =
            psystemNode->add_child("compartmentLabels");
    xmlpp::Element* compartmentsNode =
            psystemNode->add_child("membraneStructure");
    xmlpp::Element* initialMultisetsNode =
            psystemNode->add_child("initialMultisets");
    xmlpp::Element* ruleSetsNode =
            psystemNode->add_child("ruleSets");

    for(iter_t child = tree->children.begin(); 
        child != tree->children.end();
        ++child)
    {
        std::string childName(child->value.begin(),child->value.end());

        if(childName == "alphabet")
        {
            //Adding the alphabet
            addAlphabet(alphabetNode,child);
        }
	else if(childName == "initialMultisets")
	{
	    //Adding the initial multisets
	    addInitialMultisets(initialMultisetsNode,child);
	}
        else if(childName == "compartments")
	{
            //Adding compartment labels
            addCompartmentLabels(compartmentLabelsNode,child);
            //Adding membrane structure
	    addMembraneStructure(compartmentsNode,child);
        }
        else if(childName == "ruleSets")
        {
            //Adding rule sets
            addRuleSets(ruleSetsNode,child);
        }
    }

    //Writing the XML file with a formatted style
    std::string fileName = psystemName + ".xml";
    xmlFile.write_to_file_formatted(fileName);

    return fileName;
}

//Function adding the alphate to the XML specification
void ECSB::addAlphabet(xmlpp::Element* alphabetNode, iter_t alphabet)
{
    for(iter_t object = alphabet->children.begin(); 
        object != alphabet->children.end();
        ++object)
    {
        std::string objectName(object->value.begin(),object->value.end());
        objectName = removeSpaces(objectName);
        xmlpp::Element* objectNode = alphabetNode->add_child("object");
        objectNode->set_attribute("name",objectName);
    }
}

//Function adding the initial multisets to the XML specification
void ECSB::addInitialMultisets(xmlpp::Element* initialMultisetsNode,
                               iter_t initialMultisets)
{
   for(iter_t compartmentName = initialMultisets->children.begin();
       compartmentName != initialMultisets->children.end();
       ++compartmentName)
   {
       //Create a node for each initial multiset
       std::string compName(compartmentName->value.begin(),
                            compartmentName->value.end());
       compName = removeSpaces(compName);
       xmlpp::Element* initialMultisetNode =
               initialMultisetsNode->add_child("initialMultiset");
       initialMultisetNode->set_attribute("label",compName);

       //Adding the initial multiplicities for the objects in the corresponding
       //initial multiset
       for(iter_t objectName = compartmentName->children.begin();
           objectName != compartmentName->children.end();
           ++objectName)
       {
           std::string objName(objectName->value.begin(),
                               objectName->value.end());
           objName = removeSpaces(objName);
           iter_t multiplicity = objectName->children.begin();
	   std::string objMult(multiplicity->value.begin(),
                               multiplicity->value.end());
           objMult = removeSpaces(objMult);
           xmlpp::Element* objectNode = initialMultisetNode->add_child("object");
	   objectNode->set_attribute("name",objName);
           objectNode->set_attribute("multiplicity",objMult);
       }
   }
}

//Function adding the compartment labels to the XML specification
void ECSB::addCompartmentLabels(xmlpp::Element* compartmentLabelsNode,
                                iter_t compartments)
{
   for(iter_t compartmentName = compartments->children.begin();
       compartmentName != compartments->children.end();
       ++compartmentName)
   {
       std::string compName(compartmentName->value.begin(),
                            compartmentName->value.end());
       compName = removeSpaces(compName);
       xmlpp::Element* labelNode =
               compartmentLabelsNode->add_child("compartmentLabel");
       labelNode->set_attribute("label",compName);
   }
}

//Function adding the compartments to the XML specification
void ECSB::addMembraneStructure(xmlpp::Element* compartmentsNode,
                                iter_t compartments)
{
   for(iter_t compartmentName = compartments->children.begin();
       compartmentName != compartments->children.end();
       ++compartmentName)
   {
       std::string compName(compartmentName->value.begin(),
                            compartmentName->value.end());
       compName = removeSpaces(compName);
       xmlpp::Element* compartmentNode =
               compartmentsNode->add_child("compartment");
       compartmentNode->set_attribute("name",compName);
       compartmentNode->set_attribute("label",compName);

       if(!compartmentName->children.empty())
       {
           iter_t fatherCompartment = compartmentName->children.begin();
	   std::string fatherName(fatherCompartment->value.begin(),
                                  fatherCompartment->value.end());
           fatherName = removeSpaces(fatherName);
	   compartmentNode->set_attribute("father",fatherName);
       }
   }
}

//Function adding the rule sets to the XML specification
void ECSB::addRuleSets(xmlpp::Element* ruleSetsNode, iter_t ruleSets)
{
   for(iter_t compartmentName = ruleSets->children.begin();
       compartmentName != ruleSets->children.end();
       ++compartmentName)
   {
       std::string compName(compartmentName->value.begin(),
                            compartmentName->value.end());
       compName = removeSpaces(compName);
       xmlpp::Element* ruleSetNode = ruleSetsNode->add_child("ruleSet");
       ruleSetNode->set_attribute("label",compName);

       for(iter_t ruleMod = compartmentName->children.begin();
           ruleMod != compartmentName->children.end();
           ++ruleMod)
       {
           //We distinguish between module and rule using the number of
           //children in the corresponding node
           if(ruleMod->children.size() == 4)
           {
               addModule(ruleMod,ruleSetNode);
           }
           else if(ruleMod->children.size() == 7)
           {
               addRule(ruleMod,ruleSetNode);
           }
           else if(ruleMod->children.size() == 9)
           {
               addTranslRule(ruleMod,ruleSetNode);
           }

        }
   }
}

//Function adding a module to the XML specification
void ECSB::addModule(iter_t ruleMod, xmlpp::Element* ruleSetNode)
{
   //Name of the module	and addition of the corresponding node
   std::string moduleName(ruleMod->value.begin(), ruleMod->value.end());
   moduleName = removeSpaces(moduleName);
   xmlpp::Element* moduleNode = ruleSetNode->add_child("module");
   moduleNode->set_attribute("name",moduleName);

   //Details about the module
   iter_t objects = ruleMod->children.begin();
   iter_t constants = ruleMod->children.begin() + 1;
   iter_t compartments = ruleMod->children.begin() + 2;
   iter_t library = ruleMod->children.begin() + 3;

   //Adding the objects
   std::string objectsString;

   if(!objects->children.empty())
   {
       for(iter_t obj = objects->children.begin();
           obj != objects->children.end()-1;
           ++obj)
       {
           std::string objectName(obj->value.begin(),obj->value.end());
           objectsString += objectName + ",";
       }


       //Last objectName without ,
       iter_t lastObjectPointer = objects->children.end()-1;
       std::string lastObject(lastObjectPointer->value.begin(),
                              lastObjectPointer->value.end());
       objectsString += lastObject;
   }

   moduleNode->set_attribute("objects",objectsString);

   //Adding the constants
   std::string constantsString;

   if(!constants->children.empty())
   {
       for(iter_t cons = constants->children.begin();
           cons != constants->children.end()-1;
           ++cons)
       {
           std::string constantName(cons->value.begin(),cons->value.end());
           constantsString += constantName + ",";
       }

       //Last constantName without ,
       iter_t lastConstantPointer = constants->children.end()-1;
       std::string lastConstant(lastConstantPointer->value.begin(),
                                lastConstantPointer->value.end());
       constantsString += lastConstant;
   }

   moduleNode->set_attribute("constants",constantsString);

   //Adding the compartments
   std::string compartmentsString;

   if(!compartments->children.empty())
   {
       for(iter_t comp = compartments->children.begin();
           comp != compartments->children.end()-1;
           ++comp)
       {
           std::string compartmentName(comp->value.begin(),comp->value.end());
           compartmentsString += compartmentName + ",";
       }

       //Last compartmentName without ,
       iter_t lastCompartmentPointer = compartments->children.end()-1;
       std::string lastCompartment(lastCompartmentPointer->value.begin(),
                                   lastCompartmentPointer->value.end());
       compartmentsString += lastCompartment;
   }

   moduleNode->set_attribute("labels",compartmentsString);

   //Adding the library file
   std::string libraryName(library->value.begin(),library->value.end());
   libraryName = removeSpaces(libraryName);

   moduleNode->set_attribute("file",libraryName);

}
//Function adding a rule to the XML specification
void ECSB::addRule(iter_t ruleMod, xmlpp::Element* ruleSetNode)
{

    //Name of the rule and addition of the corresponding node
    std::string ruleName(ruleMod->value.begin(), ruleMod->value.end());
    ruleName = removeSpaces(ruleName);
    xmlpp::Element* ruleNode = ruleSetNode->add_child("rule");
    ruleNode->set_attribute("name",ruleName);

    //LHS
    iter_t lhs = ruleMod->children.begin() + 1;
    //LHS outside
    iter_t lhsOutside = lhs->children.begin();
    //LHS inside
    iter_t lhsInside = lhs->children.begin() + 1;
    //LHS label
    iter_t lhsLabel = lhs->children.begin() + 2;

    //Constant Name
    iter_t constantNameDef = ruleMod->children.begin() + 3;
    std::string constantName(constantNameDef->value.begin(),
                             constantNameDef->value.end());
    constantName = removeSpaces(constantName);
    ruleNode->set_attribute("constant",constantName);

    //RHS
    iter_t rhs = ruleMod->children.begin() + 5;
    //RHS outside
    iter_t rhsOutside = rhs->children.begin();
    //RHS inside
    iter_t rhsInside = rhs->children.begin() + 1;
    //RHS label
    iter_t rhsLabel = rhs->children.begin() + 2;

    //Constant
    if(ruleMod->children.size() == 7)
    {
        iter_t constant = ruleMod->children.begin() + 6;

        //Adding information about the stochastic constant
        iter_t constantValueDef = constant->children.begin() + 1;
        std::string constantValue(constantValueDef->value.begin(),
                                  constantValueDef->value.end());
        constantValue = removeSpaces(constantValue);

        ruleNode->set_attribute("value",constantValue);
    }

    //Adding information about the LHS
    xmlpp::Element* lhsNode = ruleNode->add_child("lhs");

    //LHS outside
    if(!lhsOutside->children.empty())
    {

        iter_t object = lhsOutside->children.begin();
        std::string objectName(object->value.begin(),object->value.end());
        objectName = removeSpaces(objectName);

        // Only one object can be in the outside on the lhs
        if(objectName == "+")
        {
            cerr << "error: rule " << ruleName;
            cerr << " has more than one object in the lhs outside";
            exit(EXIT_FAILURE);
        }

        xmlpp::Element* lhsOutsideObjects =
                lhsNode->add_child("listOfOutsideObjects");
        xmlpp::Element* objectNode = lhsOutsideObjects->add_child("object");
        objectNode->set_attribute("name",objectName);
    }

    //LHS inside
    if(!lhsInside->children.empty())
    {
        //Inside objects node
        xmlpp::Element* lhsInsideObjects =
                lhsNode->add_child("listOfInsideObjects");

        //Label
        std::string lhsLabelName(lhsLabel->value.begin(),lhsLabel->value.end());
        lhsLabelName = removeSpaces(lhsLabelName);
        lhsInsideObjects->set_attribute("label",lhsLabelName);

        //We distinguish if we have only one or two reactants by checking if
        //there is a + separator
        iter_t check = lhsInside->children.begin();
        std::string checkValue(check->value.begin(),check->value.end());
        checkValue = removeSpaces(checkValue);

        if(checkValue == "+")
        {
            iter_t firstObject = check->children.begin();
            iter_t secondObject = firstObject + 1;

            std::string firstObjectName(firstObject->value.begin(),
                                        firstObject->value.end());
            firstObjectName = removeSpaces(firstObjectName);
            std::string secondObjectName(secondObject->value.begin(),
                                         secondObject->value.end());
            secondObjectName = removeSpaces(secondObjectName);

            // We only support rules with two reactants
            if(secondObjectName=="+")
            {
                cerr << "error: rule " << ruleName;
                cerr << " has more than two reactants.";
                exit(EXIT_FAILURE);
            }

            xmlpp::Element* firstObjectNode =
                    lhsInsideObjects->add_child("object");
            firstObjectNode->set_attribute("name",firstObjectName);
            xmlpp::Element* secondObjectNode =
                    lhsInsideObjects->add_child("object");
            secondObjectNode->set_attribute("name",secondObjectName);
        }
        else
        {
            //There is a single object

            //Label
            std::string lhsLabelName(lhsLabel->value.begin(),
                                     lhsLabel->value.end());
            lhsLabelName = removeSpaces(lhsLabelName);
            lhsInsideObjects->set_attribute("label",lhsLabelName);

            xmlpp::Element* objectNode = lhsInsideObjects->add_child("object");
            objectNode->set_attribute("name",checkValue);
        }
    }


    //Adding information about the RHS
    xmlpp::Element* rhsNode = ruleNode->add_child("rhs");

    //RHS outside
    if(!rhsOutside->children.empty())
    {
        iter_t object = rhsOutside->children.begin();
        std::string objectName(object->value.begin(),object->value.end());
        objectName = removeSpaces(objectName);

        // Only one object can be in the outside on the lhs
        if(objectName == "+")
        {
            cerr << "error: rule " << ruleName;
            cerr << " has more than one object in the lhs outside";
            exit(EXIT_FAILURE);
        }

        xmlpp::Element* rhsOutsideObjects =
                rhsNode->add_child("listOfOutsideObjects");
        xmlpp::Element* objectNode = rhsOutsideObjects->add_child("object");
        objectNode->set_attribute("name",objectName);
    }


    //RHS inside
    if(!rhsInside->children.empty())
    {
        //Inside objects node
        xmlpp::Element* rhsInsideObjects =
                rhsNode->add_child("listOfInsideObjects");

        //Label
        std::string rhsLabelName(rhsLabel->value.begin(),rhsLabel->value.end());
        rhsLabelName = removeSpaces(rhsLabelName);
        rhsInsideObjects->set_attribute("label",rhsLabelName);

        iter_t check = rhsInside->children.begin();
        std::string checkValue(check->value.begin(),check->value.end());
        checkValue = removeSpaces(checkValue);

        //We distinguish if we have only one or two reactants by checking if
        //there is a + separator
        if(checkValue == "+")
        {
            iter_t firstObject = check->children.begin();
            iter_t secondObject = firstObject + 1;

            std::string firstObjectName(firstObject->value.begin(),
                                        firstObject->value.end());
            firstObjectName = removeSpaces(firstObjectName);
            std::string secondObjectName(secondObject->value.begin(),
                                         secondObject->value.end());
            secondObjectName = removeSpaces(secondObjectName);

            // We only support rules with two reactants
            if(secondObjectName=="+")
            {
                cerr << "error: rule " << ruleName;
                cerr << " has more than two products.";
                exit(EXIT_FAILURE);
            }

            xmlpp::Element* firstObjectNode =
                    rhsInsideObjects->add_child("object");
            firstObjectNode->set_attribute("name",firstObjectName);
            xmlpp::Element* secondObjectNode =
                    rhsInsideObjects->add_child("object");
            secondObjectNode->set_attribute("name",secondObjectName);
        }
        else
        {
            //There is a single object

            //Label
            std::string rhsLabelName(rhsLabel->value.begin(),
                                     rhsLabel->value.end());
            rhsLabelName = removeSpaces(rhsLabelName);
            rhsInsideObjects->set_attribute("label",rhsLabelName);

            xmlpp::Element* objectNode = rhsInsideObjects->add_child("object");
            objectNode->set_attribute("name",checkValue);
        }
    }
}

void ECSB::addTranslRule(iter_t ruleMod, xmlpp::Element* ruleSetNode)
{
    //Name of the rule and addition of the corresponding node
    std::string ruleName(ruleMod->value.begin(), ruleMod->value.end());
    ruleName = removeSpaces(ruleName);
    xmlpp::Element* ruleNode = ruleSetNode->add_child("rule");
    ruleNode->set_attribute("name",ruleName);

    //LHS
    iter_t lhs = ruleMod->children.begin() + 1;
    //LHS objects
    iter_t lhsObjects = lhs->children.begin();
    //LHS label
    iter_t lhsLabel = lhs->children.begin() + 1;

    //rule vector
    iter_t ruleVectorDef = lhs->children.begin() + 2;

    //Constant Name
    iter_t constantNameDef = ruleMod->children.begin() + 5;
    std::string constantName(constantNameDef->value.begin(),
                             constantNameDef->value.end());
    constantName = removeSpaces(constantName);
    ruleNode->set_attribute("constant",constantName);

    //RHS
    iter_t rhs = ruleMod->children.begin() + 7;
    //RHS objects
    iter_t rhsObjects = rhs->children.begin() + 2;

    //Constant
    if(ruleMod->children.size() == 9)
    {
        iter_t constant = ruleMod->children.begin() + 8;

        //Adding information about the stochastic constant
        iter_t constantValueDef = constant->children.begin() + 1;
        std::string constantValue(constantValueDef->value.begin(),
                                  constantValueDef->value.end());
        constantValue = removeSpaces(constantValue);

        ruleNode->set_attribute("value",constantValue);
    }

    //Adding information about the LHS
    xmlpp::Element* lhsNode = ruleNode->add_child("lhs");

    //Adding rule vector
    iter_t xDef = ruleVectorDef->children.begin();
    iter_t yDef = ruleVectorDef->children.begin() + 1;

    std::string xValue(xDef->value.begin(),xDef->value.end());
    xValue = removeSpaces(xValue);
    std::string yValue(yDef->value.begin(),yDef->value.end());
    yValue = removeSpaces(yValue);
    std::string vectorString = xValue + "," + yValue;

    ruleNode->set_attribute("vector",vectorString);

    //LHS objects
    if(!lhsObjects->children.empty())
    {
        xmlpp::Element* lhsListOfTranslocatedObjects =
                lhsNode->add_child("listOfTranslocatedObjects");

        //Label
        std::string lhsLabelName(lhsLabel->value.begin(),lhsLabel->value.end());
        lhsLabelName = removeSpaces(lhsLabelName);

        lhsListOfTranslocatedObjects->set_attribute("label",lhsLabelName);

        //Objects
        iter_t object = lhsObjects->children.begin();
        std::string objectName(object->value.begin(),object->value.end());
        objectName = removeSpaces(objectName);

        // Only one object can be translocated
        if(objectName == "+")
        {
            cerr << "error: rule " << ruleName;
            cerr << " has more than one object to translocated";
            exit(EXIT_FAILURE);
        }

        xmlpp::Element* objectNode =
                lhsListOfTranslocatedObjects->add_child("object");
        objectNode->set_attribute("name",objectName);
    }
    else
    {
        cerr << "No objects in the LHS of rule " << ruleName;
        cerr << ". No objects to translocate." ;
        exit(EXIT_FAILURE);
    }

    //Adding information about the RHS
    xmlpp::Element* rhsNode = ruleNode->add_child("rhs");

    //RHS objects
    if(!rhsObjects->children.empty())
    {
        xmlpp::Element* rhsListOfTranslocatedObjects =
                rhsNode->add_child("listOfTranslocatedObjects");

        //Label
        std::string lhsLabelName(lhsLabel->value.begin(),lhsLabel->value.end());
        lhsLabelName = removeSpaces(lhsLabelName);
        rhsListOfTranslocatedObjects->set_attribute("label",lhsLabelName);

        //Objects
        iter_t object = rhsObjects->children.begin();
        std::string objectName(object->value.begin(),object->value.end());
        objectName = removeSpaces(objectName);

        // Only one object can be translocated
        if(objectName == "+")
        {
            cerr << "error: rule " << ruleName;
            cerr << " has more than one object to translocated";
            exit(EXIT_FAILURE);
        }

        xmlpp::Element* objectNode =
                rhsListOfTranslocatedObjects->add_child("object");
        objectNode->set_attribute("name",objectName);
    }
    else
    {
        cerr << "No objects in the RHS of rule " << ruleName;
        cerr << ". No objects to translocate." ;
        exit(EXIT_FAILURE);
    }

}

unsigned long ECSB::reportErrorLine(std::string errorString, std::string fileName)
{
    
}
