/*
 * File:   spatialDistribution.cpp
 * Author: Francisco J. Romero-Campero
 *
 * Created on 15 January 2010, 09:43
 */

#include <utility>
#include <map>

#include "SpatialDistribution.h"

using namespace std;
using namespace ECSB;

//Default constructor
spatialDistribution::spatialDistribution()
{

}

//Copy constructor
spatialDistribution::spatialDistribution(const spatialDistribution& orig)
{
    name = orig.name;
    positions = orig.positions;
    numberOfCompartments = orig.numberOfCompartments;
    distribution = orig.distribution;
    noise = orig.noise;
    numberMutations = orig.numberMutations;
}

//Default destructor
spatialDistribution::~spatialDistribution()
{

}

//Constructor from xml specification
spatialDistribution::spatialDistribution(xmlpp::Node* spatialDeclaration,
                                         std::map<std::string,PSystem> pSystems)
{
  //Get the name of the spatial distribution
  xmlpp::Element* distributionElement =
          dynamic_cast<xmlpp::Element*>(spatialDeclaration);

  xmlpp::Attribute* distributionName =
          distributionElement->get_attribute("name");

  if(distributionName != NULL)
    this->name = distributionName->get_value();

  //Get the type of distribution of the positions. Fixed is the default.
  xmlpp::Attribute* distributionDeclaration =
          distributionElement->get_attribute("distribution");

  if(distributionDeclaration != NULL)
      this->distribution = distributionDeclaration->get_value();
  else
      this->distribution = "fixed";

  //Get the noise for mutation (def no noise) and number of mutations (def 10).
  xmlpp::Attribute* noiseDeclaration =
          distributionElement->get_attribute("noise");

  if(noiseDeclaration != NULL)
      this->noise = fromStringTo<double>(noiseDeclaration->get_value());
  else
      this->noise = 0;

  xmlpp::Attribute* mutationsDeclaration =
          distributionElement->get_attribute("mutation_runs");

  if(mutationsDeclaration != NULL)
      this->numberMutations =
              fromStringTo<int>(mutationsDeclaration->get_value());
  else
      this->numberMutations = 10;

  //Get the position declarations
  xmlpp::Node::NodeList positionsDeclaration =
          spatialDeclaration->get_children("position");

  //Construct the positions depending on the type of distribution and counting
  //the total number of compartments in the system
  this->numberOfCompartments = 0;

  if(this->distribution == "fixed")
    {
      //Coordinates are fixed
      fixedPositions(positionsDeclaration,pSystems);
    }
  else if(this->distribution == "random")
    {
      //Coordinates are random
      //randomPositions(positionsDeclaration,pSystems);
    }
  else
    {
      cerr << "Unknown distribution" << std::endl;
      exit(EXIT_FAILURE);
    }

}

//Function to create fixed positions
void spatialDistribution::fixedPositions(xmlpp::Node::NodeList positionsDeclaration,
        std::map<std::string,PSystem> pSystems)
{

  for(xmlpp::Node::NodeList::iterator p = positionsDeclaration.begin();
      p != positionsDeclaration.end();
      ++p)
    {
      //Get the name of the P system associated with the position
      xmlpp::Element* positionElement = dynamic_cast<xmlpp::Element*>(*p);
      std::string pSystemName =
              (positionElement->get_attribute("pSystem"))->get_value();

      //Checking if there exists such a P system
      std::map<std::string,PSystem>::iterator pSystem =
              pSystems.find(pSystemName);

      if(pSystem == pSystems.end())
      {
          cerr<< "Error: P system " << pSystemName << " not defined" << std::endl;
          exit(EXIT_FAILURE);
      }

      //Constructing the coordinates for the positions with the same P system
      std::vector< std::vector<long> > finalCoordinates = expandCoordinates(*p);

      /*
      for(std::vector< std::vector<long> >::iterator c = finalCoordinates.begin();
          c != finalCoordinates.end();
          ++c)
      {
          std::vector<long> v = *c;
          std::cout << "Pos: " << v[0] << "," << v[1] << std::endl;
          std::string pause;
          cin >> pause;
      }

       */
      //Assign the corresponding P system to the coordinates and update the
      //total number of compartments
      for(std::vector< std::vector<long> >::iterator coord = finalCoordinates.begin();
          coord != finalCoordinates.end();
          ++coord)
      {
          insertPositionPsystem(*coord, pSystem->second);
          this->numberOfCompartments += (pSystem->second).getNumberOfCompartments();
      }

    }
}

std::vector< std::vector<long> > spatialDistribution::expandCoordinates(xmlpp::Node* pos)
{
    std::vector< std::vector<long> > coordinatesPrecursor;

    //Getting the parameters associated with the positions
    std::vector<ECSB::parameter> parameters = coordinateParameters(pos);

    //Creating the coordinates
    xmlpp::Node::NodeList xCoordinateDeclaration =
            pos->get_children("Xcoordinate");
    xmlpp::Node::NodeList yCoordinateDeclaration =
            pos->get_children("Ycoordinate");

    //Getting the declaration of the coordinates
    if(xCoordinateDeclaration.empty())
    {
        cerr << "No x coordinate specified" << std::endl;
        exit(EXIT_FAILURE);
    }
    xmlpp::Node* xCoordinate = xCoordinateDeclaration.front();
    xmlpp::Element* xCoordinateElement = dynamic_cast<Element*>(xCoordinate);
    xmlpp::Attribute* xCoordinateValue =
            xCoordinateElement->get_attribute("value");

    if(yCoordinateDeclaration.empty())
    {
        cerr << "No y coordinate specified" << std::endl;
        exit(EXIT_FAILURE);
    }
    xmlpp::Node* yCoordinate = yCoordinateDeclaration.front();
    xmlpp::Element* yCoordinateElement = dynamic_cast<Element*>(yCoordinate);
    xmlpp::Attribute* yCoordinateValue =
            yCoordinateElement->get_attribute("value");

    //x coordinates precursors
    std::vector<long> xCoordinates;

    //The coordinates can be expressed as a specific value or as a range
    if(xCoordinateValue != NULL)
    {
        //The x coordinate can be expressed using a mathematical expression
        //that needs to be evaluated
        xCoordinates = evaluateExpression(xCoordinateValue->get_value(),parameters);
    }
    else //Specification as a range
    {
        long xLow = fromStringTo<long>((xCoordinateElement->get_attribute("lowerBound"))->get_value());
        long xUpper = fromStringTo<long>((xCoordinateElement->get_attribute("upperBound"))->get_value());
	for(long i =xLow; i <= xUpper; ++i)
            xCoordinates.push_back(i);
    }

    coordinatesPrecursor.push_back(xCoordinates);

    //y coordinates precursors
    std::vector<long> yCoordinates;
    if(yCoordinateValue != NULL)
    {
        yCoordinates = evaluateExpression(yCoordinateValue->get_value(),parameters);
    }
    else
    {
        long yLow = fromStringTo<long>((yCoordinateElement->get_attribute("lowerBound"))->get_value());
        long yUpper = fromStringTo<long>((yCoordinateElement->get_attribute("upperBound"))->get_value());
        for(long i =yLow; i <= yUpper; ++i)
   	    yCoordinates.push_back(i);
    }

    coordinatesPrecursor.push_back(yCoordinates);

    std::vector< std::vector<long> > finalCoordinates;

    //Depending on how the precursors were obtained we generate the final coordinates
    if((xCoordinateValue != NULL) && (yCoordinateValue != NULL))
    {
        finalCoordinates = combineParametricCoordinates(coordinatesPrecursor);
    }
    else
    {
        finalCoordinates = combineCoordinates(coordinatesPrecursor);
    }

    return finalCoordinates;
}

std::vector<ECSB::parameter> spatialDistribution::coordinateParameters(xmlpp::Node* pos)
{
    std::vector<ECSB::parameter> parameters;
    xmlpp::Node::NodeList parametersDeclaration = pos->get_children("parameter");

    for(xmlpp::Node::NodeList::iterator param = parametersDeclaration.begin();
    param != parametersDeclaration.end();
    ++param)
    {
        parameters.push_back(parameter(*param));
    }

    return parameters;
}

std::vector< std::vector<long> > spatialDistribution::combineParametricCoordinates(std::vector< std::vector<long> > coordinatesPrecursor)
{
  std::vector< std::vector<long> > combinedCoordinates;
  std::vector<long> xCoordinates = coordinatesPrecursor[0];
  std::vector<long> yCoordinates = coordinatesPrecursor[1];

  for(long i=0; i < xCoordinates.size(); ++i)
  {
      std::vector<long> coordinate;
      coordinate.push_back(xCoordinates[i]);
      coordinate.push_back(yCoordinates[i]);
      combinedCoordinates.push_back(coordinate);
  }

  return combinedCoordinates;
}

std::vector< std::vector<long> > spatialDistribution::combineCoordinates(std::vector< std::vector<long> > coordinatesPrecursor)
{
  std::vector< std::vector<long> > combinedCoordinates;
  std::vector<long> xCoordinates = coordinatesPrecursor[0];
  std::vector<long> yCoordinates = coordinatesPrecursor[1];

  for(std::vector<long>::iterator xCoord = xCoordinates.begin();
      xCoord != xCoordinates.end();
      ++xCoord)
    {
      for(std::vector<long>::iterator yCoord = yCoordinates.begin();
	  yCoord != yCoordinates.end();
	  ++yCoord)
	{
	  std::vector<long> coordinate;
	  coordinate.push_back(*xCoord);
	  coordinate.push_back(*yCoord);
	  combinedCoordinates.push_back(coordinate);
	}
    }

  return combinedCoordinates;
}

std::vector<long> spatialDistribution::evaluateExpression(std::string expression,
                                                         std::vector<ECSB::parameter> parameters)
{
    std::vector<long> result;

    //If there is no parameters nothing needs to be evaluated
    if(parameters.empty())
    {
        result.push_back(fromStringTo<int>(expression));
	return result;
    }

    //Create array containing the values for the parameters
    int numberParameters = parameters.size();
    double variables[numberParameters];

    mu::Parser parser;

    //Set the expression to evaluate
    parser.SetExpr(expression);

    //Index of the variable used for the parameter
    int index = 0;

    //Create the variables in the expression
    for(std::vector<ECSB::parameter>::iterator param = parameters.begin();
        param != parameters.end();
        ++param)
    {
        parser.DefineVar(param->getName(),&variables[index]);
        ++index;
    }

    //Generate all possible values for the parameters
    std::vector< std::vector<double> > separateValues;

    for(std::vector<ECSB::parameter>::iterator param = parameters.begin();
        param != parameters.end();
        ++param)
    {
        std::vector<double> parameterValues;

        double paramValue = param->getLowerBound();
        while(paramValue <= param->getUpperBound())
        {
            parameterValues.push_back(paramValue);
            paramValue += param->getStep();
        }

        separateValues.push_back(parameterValues);
    }

    //Combine all the possible values of the different parameters
    std::vector< std::vector<double> > values = combineValues(separateValues);

    //Evaluating the expression for all the possible values of the parameters
    for(std::vector< std::vector<double> >::iterator val = values.begin();
        val != values.end();
        ++val)
    {
        std::vector<double> currentValues = *val;
        //Instantiate the current values in the array of variables
        for(int i = 0; i < numberParameters; ++i)
        {
            variables[i] = currentValues[i];
        }

        //Evaluate the expression and add it to the result
        result.push_back(static_cast<long>(parser.Eval()));
    }

    return result;
}

std::vector< std::vector<double> > spatialDistribution::combineValues(std::vector< std::vector<double> > separateValues)
{
    std::vector< std::vector<double> > combinedValues;

    //Initialising results with vector with a single value corresponding to the
    //range for the first parameter
    std::vector<double> value;
    value.push_back(0);

    for(std::vector<double>::iterator val = (separateValues[0]).begin();
        val != (separateValues[0].end());
        ++val)
    {
        value[0]=*val;
        combinedValues.push_back(value);
    }


    for(std::vector< std::vector<double> >::iterator parameterValues = separateValues.begin()+1;
        parameterValues != separateValues.end();
        ++parameterValues)
    {

        std::vector< std::vector<double> > partialCombinedValues;

        for(std::vector< std::vector<double> >::iterator previousValues = combinedValues.begin();
            previousValues != combinedValues.end();
            ++previousValues)
        {
            for(std::vector<double>::iterator value = parameterValues->begin();
                value != parameterValues->end();
                ++value)
            {
                std::vector<double> newValues = *previousValues;
                newValues.push_back(*value);
                partialCombinedValues.push_back(newValues);
            }
        }
        combinedValues = partialCombinedValues;
    }

    return combinedValues;
}


std::vector< std::vector <std::vector<int> > > spatialDistribution::combineRandomCoordinates(std::vector< std::vector<int> > coordinatesPrecursor, int randomCopies)
{
  std::vector< std::vector<int> > combinedRandomCoordinates;
  std::vector <std::vector<int> > combinedCoordinates;

  std::vector<int> xCoordinates = coordinatesPrecursor[0];
  std::vector<int> yCoordinates = coordinatesPrecursor[1];

  for(std::vector<int>::iterator xCoord = xCoordinates.begin();
      xCoord != xCoordinates.end();
      ++xCoord)
    {
      for(std::vector<int>::iterator yCoord = yCoordinates.begin();
	  yCoord != yCoordinates.end();
	  ++yCoord)
	{
	  std::vector<int> coordinate;
	  coordinate.push_back(*xCoord);
	  coordinate.push_back(*yCoord);
	  combinedCoordinates.push_back(coordinate);
	}
    }

  //Pick random coordinates
  srand((unsigned)time(0));

  std::vector< std::vector<int> >::iterator coordinatesIterator = combinedCoordinates.begin();
  for(int k = 0; k < randomCopies; ++k)
    {
      int numberCoordinates = combinedCoordinates.size();
      int randomPosition = rand()%numberCoordinates;
      combinedRandomCoordinates.push_back(combinedCoordinates[randomPosition]);
      combinedCoordinates.erase(coordinatesIterator+randomPosition);
    }

  std::vector< std::vector <std::vector<int> > > coordinatesFinal;
  coordinatesFinal.push_back(combinedRandomCoordinates);
  coordinatesFinal.push_back(combinedCoordinates);
  return coordinatesFinal;
}

void spatialDistribution::addNoise(double probability, long xLowerValue,
                                   long xUpperValue, long yLowerValue, long yUpperValue)
{
    //Initialise seed
    srand (time(NULL));

    std::map<std::vector<long>,std::string > mutatedPositions;
    std::map<std::vector<long>,std::string > mapPositions = positions;

    //Add the borders without mutations
    for(long coord = xLowerValue; coord <= xUpperValue; ++coord)
    {
        //Lower border
        std::vector<long> position1;
        position1.push_back(coord);
        position1.push_back(yLowerValue);

        mutatedPositions[position1] = mapPositions[position1];
        mapPositions.erase(position1);

        //Upper border
        std::vector<long> position2;
        position2.push_back(coord);
        position2.push_back(yUpperValue);

        mutatedPositions[position2] = mapPositions[position2];
        mapPositions.erase(position2);
    }

    for(long coord = yLowerValue + 1; coord <= yUpperValue - 1; ++coord)
    {
        //Left border
        std::vector<long> position1;
        position1.push_back(xLowerValue);
        position1.push_back(coord);

        mutatedPositions[position1] = mapPositions[position1];
        mapPositions.erase(position1);

        //Right border
        std::vector<long> position2;
        position2.push_back(xUpperValue);
        position2.push_back(coord);

        mutatedPositions[position2] = mapPositions[position2];
        mapPositions.erase(position2);
    }


    //Adding noise to the remaining positions
    while(!mapPositions.empty())
    {
        std::map<std::vector<long>,std::string >::iterator mapPos = mapPositions.begin();
        std::vector<long> pos1 = mapPos->first;
        std::string pSystem1 = mapPos->second;

        double randomNumber = (double)rand()/RAND_MAX;

        if (randomNumber <= probability)
        {
            std::vector<long> pos2 = neighbouringPosition(pos1);
            std::map<std::vector<long>,std::string>::iterator neigh = mapPositions.find(pos2);

            if(neigh == mapPositions.end())
            {
                mapPositions.erase(pos1);
                mutatedPositions[pos1] = pSystem1;
            }
            else
            {
                //swap P systems
                std::string pSystem2 = neigh->second;

                mutatedPositions[pos1] = pSystem2;
                mutatedPositions[pos2] = pSystem1;

                //Remove automata from the map
                mapPositions.erase(pos1);
                mapPositions.erase(pos2);
            }
        }
        else
        {
            mapPositions.erase(pos1);
            mutatedPositions[pos1] = pSystem1;
        }
    }

    this->positions = mutatedPositions;
}

std::vector<long> spatialDistribution::neighbouringPosition(std::vector<long> pos)
{
    std::vector<long> result;

    double randomNumber = (double)rand()/RAND_MAX;

    if (randomNumber <= 0.25)
    {
        result.push_back(pos[0] + 1);
        result.push_back(pos[1] + 1);
    }
    else if ((randomNumber > 0.25) && (randomNumber <= 0.5))
    {
        result.push_back(pos[0] - 1);
        result.push_back(pos[1] + 1);
    }
    else if ((randomNumber > 0.5) && (randomNumber <= 0.75))
    {
        result.push_back(pos[0] + 1);
        result.push_back(pos[1] - 1);
    }
    else if ((randomNumber > 0.75) && (randomNumber <= 1))
    {
        result.push_back(pos[0] - 1);
        result.push_back(pos[1] - 1);
    }

    return result;
}

void spatialDistribution::show()
{
    for(std::map<vector<long>,std::string>::iterator pos = this->positions.begin();
    pos != this->positions.end();
    ++pos)
    {
        std::vector<long> position = pos->first;
        std::cout << "Position (" << position[0] << "," << position[1];
        std::cout << ") P system " << pos->second << std::endl;

    }
}