// create parameter classes from parameter set xml
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#include <iostream>
#include <algorithm>
#include <cctype>
#include <string>
#include <libxml++/libxml++.h>
#include <stdio.h>
#include <stdlib.h>

using namespace std;
using namespace xmlpp;

// default filename base for output header and cpp files
#define DEFAULT_BASE "Parameters"

// file pointers for output header and cpp files
FILE *headerfp;
FILE *cppfp;

typedef struct Parameter {
	string *pName;
	string *pType;
	string *pCType;
	string *pDescription;
	string *pValue;
	string *pDefault;
	string *pUbound;
	string *pLbound;
	string *printfFlag;
};

void processParameterSet(Node *psNode)
{
	Node *node;
	Node::NodeList nodeList;
	Node::NodeList::iterator iter;
	Parameter *ps;
	string psClassName, pName, pType, pCType, pDescription, pValue, pDefault;
	string pUbound, pLbound;
	unsigned int index, num_parameters;
	string::size_type pos;
	
	// get parameter set name
	psClassName = (dynamic_cast<Element *>(psNode))->get_attribute_value("name");
	if(psClassName.empty()) {
		cerr << "error: no name attribute specified for parameterSet\n";
		exit(EXIT_FAILURE);
	}
	
	// get list of parameters
	nodeList = psNode->get_children("parameter");
	if(nodeList.empty()) {
		cerr << "error: no parameterSet elements specified\n";
		exit(EXIT_FAILURE);
	}
	
	// allocate memory to store parameters
	if(!(ps = (Parameter *) calloc(nodeList.size(), sizeof(Parameter)))) {
		cerr << "error: couldn't allocate memory for parameter set\n";
		exit(EXIT_FAILURE);
	}
	
	// parse each parameter
	num_parameters = nodeList.size();
	index = 0;
	for(iter = nodeList.begin(); iter != nodeList.end(); ++iter) {
		node = *iter;
		
		// get parameter name
		ps[index].pName = new string( \
				(dynamic_cast<Element *>(node))->get_attribute_value("name"));
		if(ps[index].pName->empty()) {
			cerr << "error: no name attribute specified for parameter\n";
			exit(EXIT_FAILURE);
		}
		
		// get parameter type
		ps[index].pType = new string ( \
				(dynamic_cast<Element *>(node))->get_attribute_value("type"));
		if(ps[index].pType->empty()) {
			cerr << "error: no type attribute specified for parameter " \
					<< ps[index].pName << "\n";
			exit(EXIT_FAILURE);
		}
		
		// get parameter description
		ps[index].pDescription = new string((dynamic_cast<Element *>(node))-> \
				get_attribute_value("description"));
		
		// get parameter default
		ps[index].pDefault = new string((dynamic_cast<Element *>(node))-> \
				get_attribute_value("default"));
		
		// get parameter value
		ps[index].pValue = new string((dynamic_cast<Element *>(node))-> \
				get_attribute_value("value"));
		
		// get parameter upper bound
		ps[index].pUbound = new string((dynamic_cast<Element *>(node))-> \
				get_attribute_value("ubound"));
		
		// get parameter lower bound
		ps[index].pLbound = new string((dynamic_cast<Element *>(node))-> \
				get_attribute_value("lbound"));
		
		// determine c type for parameter
		pType = *ps[index].pType;
		if(pType == "int" || pType == "float" || pType == "double" || \
				pType == "char" || pType == "bool" || pType == "long") {
			ps[index].pCType = new string(pType);
		} else {
			if(pType == "string" || pType == "filename") {
				ps[index].pCType = new string("char");
			} else {
				if(pType == "cppstring") {
					ps[index].pCType = new string("std::string");
				} else {
					if(pType == "ulong") {
						ps[index].pCType = new string("unsigned long");
					} else {
						if(pType == "uint") {
							ps[index].pCType = new string("unsigned int");
						} else {
							cerr << "error: invalid type " << pType << \
									" for parameter " << ps[index].pName << "\n";
							exit(EXIT_FAILURE);
						}
					}
				}
			}
		}
		index++;
	}
	
	// write class preamble
	fprintf(headerfp, "\n// %s class\n", psClassName.c_str());
	fprintf(headerfp, "class %s : public ECSB::ParameterSet {\n", \
			psClassName.c_str());
	fprintf(headerfp, "\tpublic:\n");
	fprintf(headerfp, "\t\t// constructor\n");
	fprintf(headerfp, "\t\t%s(void);\n\n", psClassName.c_str());
	fprintf(headerfp, "\t\t// parameters\n");
	
	// write class declarations
	for(index = 0; index < num_parameters; index++) {
		if(!ps[index].pDescription->empty())
			fprintf(headerfp, "\t\t// %s\n", ps[index].pDescription->c_str());
		if(*ps[index].pType == "string" || *ps[index].pType == "filename")
			fprintf(headerfp, "\t\t%s %s[PARAMETER_STR_MAX];\n", \
					ps[index].pCType->c_str(), ps[index].pName->c_str());
		else
			fprintf(headerfp, "\t\t%s %s;\n", ps[index].pCType->c_str(), \
					ps[index].pName->c_str());
	}
	fprintf(headerfp, "\n\t\t// set a parameter value\n");
	fprintf(headerfp, "\t\tbool setParameter(char *name, char *value);\n");
		
	fprintf(headerfp, "\n\t\t// write parameter set to file\n");
	fprintf(headerfp, "\t\tvoid write(const char *filename);\n");
		
	// write class postamble
	fprintf(headerfp, "};\n");
	
	// write constructor preamble
	fprintf(cppfp, "\n// constructor\n");
	fprintf(cppfp, "%s::%s(void) : ParameterSet(\"%s\")\n{\n", \
			psClassName.c_str(), psClassName.c_str(), psClassName.c_str());
	fprintf(cppfp, "\t// set initial values\n");
			
	// write constructor definitions
	for(index = 0; index < num_parameters; index++) {
		if(!ps[index].pDefault->empty()) {
			if(*ps[index].pType == "cppstring") {
				fprintf(cppfp, "\t%s = \"%s\";\n", ps[index].pName->c_str(), \
						ps[index].pDefault->c_str());
			} else {
				if(*ps[index].pType == "string" || \
						*ps[index].pType == "filename") {
					fprintf(cppfp, \
							"\tstrncpy(%s, \"%s\", PARAMETER_STR_MAX);\n", \
							ps[index].pName->c_str(), \
							ps[index].pDefault->c_str());
				} else {
					if(*ps[index].pType == "char") {
						fprintf(cppfp, "\t%s = '%s';\n", \
								ps[index].pName->c_str(), \
								ps[index].pDefault->c_str());
					} else {
						fprintf(cppfp, "\t%s = %s;\n", \
								ps[index].pName->c_str(), \
								ps[index].pDefault->c_str());
					}
				}
			}
		}
	}
	
	// write constructor postamble
	fprintf(cppfp, "}\n");
	
	// write setParameter method preamble
	fprintf(cppfp, "\n// set a parameter value\n");
	fprintf(cppfp, "bool %s::setParameter(char *name, char *value)\n{\n", \
			psClassName.c_str());
			
	// write setParameter method
	for(index = 0; index < num_parameters; index++) {
		fprintf(cppfp, "\tif(strcmp(name, \"%s\") == 0) {\n", \
				ps[index].pName->c_str());			
		if(*ps[index].pType == "cppstring") {
			fprintf(cppfp, "\t\t%s = value;\n", ps[index].pName->c_str());
			ps[index].printfFlag = new string("s");
			goto done;
		}
		if(*ps[index].pType == "string") {
			pValue = *ps[index].pValue;
			if(!pValue.empty()) {
				fprintf(cppfp, "\t\tif(!(");
				pos = 0;
				do {
					if(pos)
						fprintf(cppfp, " || ");
					fprintf(cppfp, "strcmp(value, \"%s\") == 0", \
							pValue.substr(pos, pValue.find('|', pos) - \
									pos).c_str());
					pos = pValue.find('|', pos);
					if(pos != string::npos)
						pos++;
				} while (pos != string::npos);
				fprintf(cppfp, ")) {\n");				
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: invalid value" \
						" \\\"\" << value << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\"\\\" for parameter %s\\n\";\n", \
						ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\tstrncpy(%s, value, PARAMETER_STR_MAX);\n", \
					ps[index].pName->c_str());
			fprintf(cppfp, "\t\t%s[PARAMETER_STR_MAX - 1] = '\\0';\n", \
					ps[index].pName->c_str());
			ps[index].printfFlag = new string("s");
			goto done;
		}
		if(*ps[index].pType == "filename") {
			fprintf(cppfp, "\t\tchar path[PARAMETER_STR_MAX];\n");
			fprintf(cppfp, "\t\tstrncpy(path, value, PARAMETER_STR_MAX);\n");
			fprintf(cppfp, "\t\tchar *path2 = dirname(path);\n");
			fprintf(cppfp, "\t\tif(access(path2, F_OK) != 0) {\n");
			fprintf(cppfp, "\t\t\tstd::cerr << \"error: unable to" 
					" access file \" << value << \"\\n\";\n");
			fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			fprintf(cppfp, "\t\tstrncpy(%s, value, PARAMETER_STR_MAX);\n", \
					ps[index].pName->c_str());
			fprintf(cppfp, "\t\t%s[PARAMETER_STR_MAX - 1] = '\\0';\n", \
					ps[index].pName->c_str());
			ps[index].printfFlag = new string("s");
			goto done;
		}
		if(*ps[index].pType == "char") {
			fprintf(cppfp, "\t\t%s = *value;\n", ps[index].pName->c_str());
			ps[index].printfFlag = new string("c");
			goto done;
		}
		if(*ps[index].pType == "double") {
			if(!ps[index].pUbound->empty()) {
				fprintf(cppfp, "\t\tif(atof(value) > %f) {\n", \
						atof(ps[index].pUbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atof(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			if(!ps[index].pLbound->empty()) {
				fprintf(cppfp, "\t\tif(atof(value) < %f) {\n", \
						atof(ps[index].pLbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atof(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\t%s = atof(value);\n", ps[index].pName->c_str());
			ps[index].printfFlag = new string("f");
			goto done;
		}
		if(*ps[index].pType == "float") {
			if(!ps[index].pUbound->empty()) {
				fprintf(cppfp, "\t\tif(atof(value) > %f) {\n", \
						atof(ps[index].pUbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atof(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			if(!ps[index].pLbound->empty()) {
				fprintf(cppfp, "\t\tif(atof(value) < %f) {\n", \
						atof(ps[index].pLbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atof(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\t%s = (float) atof(value);\n", \
					ps[index].pName->c_str());
			ps[index].printfFlag = new string("f");
			goto done;
		}
		if(*ps[index].pType == "bool") {
			fprintf(cppfp, "\t\tif(strcmp(value, \"true\") == 0 " \
					"|| *value == '1')\n");
			fprintf(cppfp, "\t\t\t%s = true;\n", ps[index].pName->c_str());
			fprintf(cppfp, "\t\telse\n");
			fprintf(cppfp, "\t\t\t%s = false;\n", ps[index].pName->c_str());
			ps[index].printfFlag = new string("d");
			goto done;
		}
		if(*ps[index].pType == "int") {
			if(!ps[index].pUbound->empty()) {
				fprintf(cppfp, "\t\tif(atoi(value) > %d) {\n", \
						atoi(ps[index].pUbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atoi(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			if(!ps[index].pLbound->empty()) {
				fprintf(cppfp, "\t\tif(atoi(value) < %d) {\n", \
						atoi(ps[index].pLbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atoi(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\t%s = atoi(value);\n", ps[index].pName->c_str());
			ps[index].printfFlag = new string("d");
			goto done;
		}
		if(*ps[index].pType == "long") {
			if(!ps[index].pUbound->empty()) {
				fprintf(cppfp, "\t\tif(atol(value) > %ld) {\n", \
						atol(ps[index].pUbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atol(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			if(!ps[index].pLbound->empty()) {
				fprintf(cppfp, "\t\tif(atol(value) < %ld) {\n", \
						atol(ps[index].pLbound->c_str()));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << atol(value) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\t%s = atol(value);\n", \
					ps[index].pName->c_str());
			ps[index].printfFlag = new string("ld");
			goto done;
		}
		if(*ps[index].pType == "ulong") {
			if(!ps[index].pUbound->empty()) {
				fprintf(cppfp, "\t\tif(atol(value) > %lu) {\n", \
						strtoul(ps[index].pUbound->c_str(), NULL, 10));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" <<strtoul(value, NULL, 10) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			if(!ps[index].pLbound->empty()) {
				fprintf(cppfp, "\t\tif(strtoul(value, NULL, 10) < %lu) {\n", \
						strtoul(ps[index].pLbound->c_str(), NULL, 10));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << strtoul(value, NULL, 10) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\t%s = strtoul(value, NULL, 10);\n", \
					ps[index].pName->c_str());
			ps[index].printfFlag = new string("ld");
			goto done;
		}
		if(*ps[index].pType == "uint") {
			if(!ps[index].pUbound->empty()) {
				fprintf(cppfp, "\t\tif(atol(value) > %lu) {\n", \
						strtoul(ps[index].pUbound->c_str(), NULL, 10));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" <<strtoul(value, NULL, 10) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			if(!ps[index].pLbound->empty()) {
				fprintf(cppfp, "\t\tif(strtoul(value, NULL, 10) < %lu) {\n", \
						strtoul(ps[index].pLbound->c_str(), NULL, 10));
				fprintf(cppfp, "\t\t\tstd::cerr << \"error: value" \
						" \" << strtoul(value, NULL, 10) << \\\n");
				fprintf(cppfp, "\t\t\t\t\t\" out of range for parameter " \
						"%s\\n\";\n", ps[index].pName->c_str());
				fprintf(cppfp, "\t\t\texit(EXIT_FAILURE);\n\t\t}\n");
			}
			fprintf(cppfp, "\t\t%s = (unsigned int) strtoul(value, NULL, 10);\n", \
					ps[index].pName->c_str());
			ps[index].printfFlag = new string("u");
			goto done;
		}
		done:
			fprintf(cppfp, "\t\treturn true;\n\t}\n");			
	}
					
	// write setParameter method postamble
	fprintf(cppfp, "\n\treturn false;\n}\n");			
		
	// write write method preamble
	// write setParameter method preamble
	fprintf(cppfp, "\n// write parameter set to file\n");
	fprintf(cppfp, "void %s::write(const char *filename)\n{\n", \
			psClassName.c_str());
	fprintf(cppfp, "\tFILE *fp;\n");
	fprintf(cppfp, "\n\t// open output file\n");
	fprintf(cppfp, "\tif(!(fp = fopen(filename, \"w\"))) {\n");
	fprintf(cppfp, "\t\tstd::cerr << \"error: unable to open file " \
			"\" << filename << \"\\n\";\n");
	fprintf(cppfp, "\t\texit(EXIT_FAILURE);\n\t}\n\n");
			
	// write write method
	// write xml header
	fprintf(cppfp, "\t// write xml header\n");
	fprintf(cppfp, "\tfprintf(fp, \"<?xml version=\\\"1.0\\\" " \
			"encoding=\\\"utf-8\\\" ?>\\n\");\n");
	fprintf(cppfp, "\n\tfprintf(fp, \"<parameters " \
			"xmlns=\\\"http://www.cpib.ac.uk/~jpt\\\"\\n\");\n");
	fprintf(cppfp, "\tfprintf(fp, \"\\txmlns:xsi=\\\"" \
			"http://www.w3.org/2001/XMLSchema-instance\\\"\\n\");\n");
	fprintf(cppfp, "\tfprintf(fp, \"\\txsi:schemaLocation=\\\"" \
			"http://www.cpib.ac.uk/~jpt parameter.xsd\\\">\\n\");\n");

	// write parameter values
	fprintf(cppfp, "\n\t// write parameter values\n");
	fprintf(cppfp, "\tfprintf(fp, \"\\n<!-- parameter definitions -->\\n\"" \
			");\n");
	fprintf(cppfp, "\tfprintf(fp, \"<parameterSet name=\\\"%s\\\">\\n\")" \
			";\n", psClassName.c_str());
	for(index = 0; index < num_parameters; index++) {
		fprintf(cppfp, "\tfprintf(fp, " \
				"\"\\t<parameter name=\\\"%s\\\" value=\\\"%%%s\\\"/>\\n\", "
				"%s", ps[index].pName->c_str(), \
				ps[index].printfFlag->c_str(), ps[index].pName->c_str());
		if(*ps[index].pType == "cppstring")
			fprintf(cppfp, ".c_str()");
		fprintf(cppfp, ");\n");
	}
	fprintf(cppfp, "\tfprintf(fp, \"</parameterSet>\\n\");\n");

	fprintf(cppfp, "\n\t// write xml trailer\n");
	fprintf(cppfp, "\tfprintf(fp, \"\\n</parameters>\\n\");\n");
		
	fprintf(cppfp, "\n\t// close output file\n");
	fprintf(cppfp, "\tfclose(fp);\n}\n");
		
	// free memory for parameter set
	free(ps);
}

int main(int argc, char **argv)
{
	Node::NodeList nodeList;
	Node::NodeList::iterator iter;
	char *infile;
	string outbase, filename;
	
	// parse command line
	if(argc != 2) {
		cerr << "usage: ecsb-make-parameter-class XML-FILE\n";
		exit(EXIT_FAILURE);
	}
	infile = argv[1];

	// load xml document
	DomParser parser(infile);
	if(!parser) {
		cerr << "error: couldn't load parameter file " << infile << "\n";
		exit(EXIT_FAILURE);
	}
	Document *document = parser.get_document();
	
	// check root node
	if(document->get_root_node()->get_name() != "parameters") {
		cerr << "error: no parameters element\n";
		exit(EXIT_FAILURE);
	}
	
	// determine if output filename specified in xml
	outbase = (dynamic_cast<Element *>(document->get_root_node()))-> \
			get_attribute_value("name");
	if(outbase.empty())
		outbase = "Parameters";
	
	// create output header and cpp files
	filename = outbase + ".h";
	if(!(headerfp = fopen(filename.c_str(), "w"))) {
		cerr << "error: couldn't create output file " << filename << "\n";
		exit(EXIT_FAILURE);
	}
	filename = outbase + ".cpp";
	if(!(cppfp = fopen(filename.c_str(), "w"))) {
		cerr << "error: couldn't create output file " << filename << "\n";
		exit(EXIT_FAILURE);
	}

	// write header preamble
	fprintf(headerfp, "// parameter class definitions\n");
	fprintf(headerfp, "// automatically generated by make-parameter-class\n");
	fprintf(headerfp, "// part of libecsb - the executable cell systems biology " \
			"library\n");
	fprintf(headerfp, "// copyright 2008, 2009 jamie twycross, " \
			"jpt AT cs.nott.ac.uk\n");
	fprintf(headerfp, "// released under GNU GPL version 3\n\n");
	filename = outbase + "_h";
	transform(filename.begin(), filename.end(), filename.begin(), \
			(int(*)(int)) toupper);
	fprintf(headerfp, "#ifndef %s\n", filename.c_str());
	fprintf(headerfp, "#define %s\n\n", filename.c_str());
	fprintf(headerfp, "#include <ecsb/ecsb.h>\n");

	// write cpp preamble
	fprintf(cppfp, "// parameter class definitions\n");
	fprintf(cppfp, "// automatically generated by make-parameter-class\n");
	fprintf(cppfp, "// part of libecsb - the executable cell systems biology " \
			"library\n");
	fprintf(cppfp, "// copyright 2008, 2009 jamie twycross, " \
			"jpt AT cs.nott.ac.uk\n");
	fprintf(cppfp, "// released under GNU GPL version 3\n\n");
	fprintf(cppfp, "#include \"%s.h\"\n\n", outbase.c_str());
	fprintf(cppfp, "#include <string.h>\n");
	fprintf(cppfp, "#include <stdlib.h>\n");
	fprintf(cppfp, "#include <iostream>\n");
	fprintf(cppfp, "#include <libgen.h>\n");
	fprintf(cppfp, "#include <unistd.h>\n");

	// get list of parameter sets
	nodeList = document->get_root_node()->get_children("parameterSet");
	if(nodeList.empty()) {
		cerr << "error: no parameterSet elements specified\n";
		exit(EXIT_FAILURE);
	}
	
	// process each parameter set
	for(iter = nodeList.begin(); iter != nodeList.end(); ++iter) {
		processParameterSet(*iter);
	}	
		
	// write header postamble
	fprintf(headerfp, "\n#endif\n");
	
	// close header and cpp files
	fclose(cppfp);
	fclose(headerfp);
		
	return 0;
}
